# -*- coding: utf-8 -*-
"""
The :mod:`.utility` module has several functions that support celloracle.

"""

from .make_log import makelog

from .utility import (save_as_pickled_object, load_pickled_object,
                      intersect,
                      exec_process,
                      standard, inverse_dictionary,
                      adata_to_color_dict,
                      transfer_all_colors_between_anndata,
                      transfer_color_between_anndata,
                      knn_data_transferer,
                      update_adata)
from .load_hdf5 import load_hdf5

from .pandas_utility_for_jupyternotebook import init_datatable_mode

from .package_version_checker import check_python_requirements
__all__ = [
           "makelog",
           "save_as_pickled_object",
           "load_pickled_object",
           "intersect",
           "exec_process",
           "standard",
           "load_hdf5",
           "inverse_dictionary"
           "adata_to_color_dict",
           "transfer_all_colors_between_anndata",
           "transfer_color_between_anndata",
           "knn_data_transferer",
           "update_adata"
           ]



# -*- coding: utf-8 -*-



import io
import logging
import os


from datetime import datetime


class makelog():
    """
    This is a class for making log.

    """
    def __init__(self, file_name=None, directory=None):

        name = datetime.now().ctime()

        # ログの出力名を設定
        logger = logging.Logger('LoggingTest1')

        # ログレベルの設定
        logger.setLevel(10)

        # ログのファイル出力先を設定

        if not file_name is None:
            name = file_name + " " + name

        file_path = f'log {name}.log'

        if not directory is None:
            os.makedirs(directory, exist_ok=True)
            file_path = os.path.join(directory, file_path)

        fh = logging.FileHandler(file_path)
        logger.addHandler(fh)

        # ログのコンソール出力の設定
        sh = logging.StreamHandler()
        logger.addHandler(sh)

        # ログの出力形式の設定
        formatter = logging.Formatter('%(asctime)s:%(lineno)d:%(levelname)s:%(message)s')
        fh.setFormatter(formatter)
        sh.setFormatter(formatter)

        self.logger = logger

    def info(self, comment):
        """
        Add comment into the log file.

        Args:
            comment (str): comment.
            
        """
        self.logger.log(20, comment)



# -*- coding: utf-8 -*-
'''
This is a series of custom functions for the inferring of GRN from single cell RNA-seq data.

Codes were written by Kenji Kamimoto.


'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing


import os, warnings
import subprocess
import pandas as pd
from joblib import dump, load

from multiprocessing import cpu_count



from ..utility import exec_process
from ..network_analysis import __path__ as parent_path
#import seaborn as sns

config = {"R_path": "R", "n_cpu": cpu_count()}

def get_R_path():
    r = subprocess.check_output("which R".split(" ")).decode()
    r = r.replace("\n", "")
    return r

try:
    config["R_path"] = get_R_path()
except:
    pass
    #print("R path not found.")

def set_R_path(R_path):
    if not R_path.endswith("R"):
        raise ValueError("R_path should end with R.")

    command = f"{R_path} --version"
    try:
        print("Testing R...\n")
        exec_process(command, message=True)

        config["R_path"] = R_path
    except:
        raise ValueError(f"Could not find R at {R_path}.")


def test_R_libraries_installation(show_all_stdout=False):
    warnings.warn(
        "\ntest_R_libraries_installation network analysis function does not require R package anymore after version 0.10.0.\n" + \
        "Although the old network analysis functions are still available now, they will be removed soon.",
        DeprecationWarning
    )

def _check_R_libraries_installation(show_all_stdout=False):

    """
    CellOracle.network_analysis use several R libraries for network analysis.
    This is a test function to check for instalation of the necessary R libraries.
    """

    result = True
    print("R path: " + config["R_path"])
    r_libraries = ["igraph", "linkcomm", "rnetcarto"]#,"gProfileR"]
    for i in r_libraries:
        try:
            if show_all_stdout:
                command = f"{config['R_path']} --file={parent_path[0]}/rscripts_for_network_analysis/test_{i}.R"
            else:
                command = f"{config['R_path']} --slave --file={parent_path[0]}/rscripts_for_network_analysis/test_{i}.R"
            exec_process(command, message=True)
            print(f"checking R library installation: {i} -> OK")
        except:
            print(f"checking_installation: {i} -> NG")
            print(f" R library, {i} is unavailable. Please check installation.")
            result = False
    return result



def _get_cartography_by_Rscripts_inparallel(dict_links, id_dict, output_folder="network_analysis",
                      message=False, n_parallel=-1):

    if n_parallel == -1:
        n_parallel = config["n_cpu"]
    li = list(dict_links.keys())
    N = len(li)


    def internal(li_):
        process_list = []
        for i in li_:
            folder = os.path.join(output_folder, str(id_dict[i]))
            os.makedirs(folder, exist_ok=True)

            link_path =folder + "/linkList.csv"
            dict_links[i][["source", "target", "coef_abs"]].to_csv(link_path, index=None)

            r_dir = config["R_path"][:-1]

            command = f"{r_dir}Rscript {parent_path[0]}/rscripts_for_network_analysis/get_cartography.R {folder}"
            #print(command)
            process_list.append(exec_process(command, message=message,
                                             wait_finished=False, return_process=True))

        for process, name in zip(process_list, li_):

            process.wait()
            if process.returncode != 0:
                print(f'{name}: Build process aborts.')
            else:
                print(f'{name}: finished.')

    if N%n_parallel == 0:
        for k in range(N//n_parallel):
            print(f"processing... batch {k+1}/{N//n_parallel}")
            sub_list = li[n_parallel*k:n_parallel*(k+1)]
            internal(sub_list)

    else:
        for k in range(N//n_parallel):
            print(f"processing... batch {k+1}/{(N//n_parallel)+1}")
            sub_list = li[n_parallel*k:n_parallel*(k+1)]
            internal(sub_list)

        print(f"processing... batch {(N//n_parallel)+1}/{(N//n_parallel)+1}")
        sub_list = li[(N//n_parallel)*n_parallel: N]
        internal(sub_list)


def _get_network_score_by_Rscripts_inparallel(dict_links, id_dict, output_folder="network_analysis",
                      message=False, n_parallel=-1):

    if n_parallel == -1:
        n_parallel = config["n_cpu"]
    li = list(dict_links.keys())
    N = len(li)


    def internal(li_):
        process_list = []
        for i in li_:
            folder = os.path.join(output_folder, str(id_dict[i]))
            os.makedirs(folder, exist_ok=True)

            link_path =folder + "/linkList.csv"
            dict_links[i][["source", "target", "coef_abs"]].to_csv(link_path, index=None)

            r_dir = config["R_path"][:-1]

            command = f"{r_dir}Rscript {parent_path[0]}/rscripts_for_network_analysis/get_newtork_scores.R {folder}"

            process_list.append(exec_process(command, message=message,
                                             wait_finished=False, return_process=True))

        for process, name in zip(process_list, li_):

            process.wait()
            if process.returncode != 0:
                print(f'{name}: Build process aborts.')
            else:
                print(f'{name}: finished.')

    if N%n_parallel == 0:
        for k in range(N//n_parallel):
            print(f"processing... batch {k+1}/{N//n_parallel}")
            sub_list = li[n_parallel*k:n_parallel*(k+1)]
            internal(sub_list)

    else:
        for k in range(N//n_parallel):
            print(f"processing... batch {k+1}/{(N//n_parallel)+1}")
            sub_list = li[n_parallel*k:n_parallel*(k+1)]
            internal(sub_list)

        print(f"processing... batch {(N//n_parallel)+1}/{(N//n_parallel)+1}")
        sub_list = li[(N//n_parallel)*n_parallel: N]
        internal(sub_list)



# -*- coding: utf-8 -*-


import os, sys

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from ..trajectory.oracle_utility import _adata_to_color_dict


def visualize_developmental_analysis_ver1(self, scale_for_pseudotime=30, scale_for_simulated=30, s=10, s_grid=30, vmin=-1, vmax=1):

    original_embedding=self.oracle_dev.embedding
    original_value=self.oracle_dev.pseudotime
    mass_filter=self.oracle_dev.mass_filter
    grid=self.oracle_dev.flow_grid
    value_on_grid=self.oracle_dev.new_pseudotime
    gradient_pseudotime=self.oracle_dev.gradient
    gradient_simulated=self.oracle_dev.flow
    inner_product=self.oracle_dev.inner_product


    fig, ax = plt.subplots(2, 4, figsize=[20,10])

    ax[0, 0].scatter(original_embedding[:, 0], original_embedding[:, 1], c=original_value, cmap="rainbow", s=s)
    ax[0, 0].set_title("Pseudotime")
    ax[0,0].axis("off")

    ax[0, 1].scatter(grid[~mass_filter, 0], grid[~mass_filter, 1], c=value_on_grid[~mass_filter], cmap="rainbow", s=s_grid)
    ax[0, 1].set_title("Pseudotime on grid")
    ax[0, 1].axis("off")

    ax[0, 2].quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_pseudotime[~mass_filter, 0], gradient_pseudotime[~mass_filter, 1],
                    scale=scale_for_pseudotime)
    ax[0, 2].set_title("Gradient of pseudotime \n(=Development flow)")
    ax[0, 2].axis("off")



    ax[0, 3].scatter(original_embedding[:, 0], original_embedding[:, 1], c=original_value, cmap="rainbow", s=s)
    ax[0, 3].quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_pseudotime[~mass_filter, 0], gradient_pseudotime[~mass_filter, 1],
                    scale=scale_for_pseudotime)
    ax[0, 3].set_title("Pseudotime + \nDevelopment flow")
    ax[0, 3].axis("off")

    ax[1, 0].quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_simulated[~mass_filter, 0], gradient_simulated[~mass_filter, 1],
                    scale=scale_for_simulated, zorder=20000)
    ax[1, 0].set_title("Perturb simulation")
    ax[1, 0].axis("off")


    ax[1, 1].scatter(grid[~mass_filter, 0], grid[~mass_filter, 1], c=inner_product[~mass_filter], cmap="coolwarm", s=s_grid, vmin=vmin, vmax=vmax)
    ax[1, 1].set_title("Inner product of \n Perturb simulation * Development flow")
    ax[1, 1].axis("off")


    ax[1, 2].scatter(grid[~mass_filter, 0], grid[~mass_filter, 1], c=inner_product[~mass_filter], cmap="coolwarm", s=s_grid, vmin=vmin, vmax=vmax)
    ax[1, 2].quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_simulated[~mass_filter, 0], gradient_simulated[~mass_filter, 1],
                    scale=scale_for_simulated, zorder=20000)
    ax[1, 2].axis("off")
    ax[1, 2].set_title("Inner product of \n Perturb simulation * Development flow \n + Perturb simulation")

    ax_ = ax[1, 3]
    pcm = ax_.scatter(value_on_grid[~mass_filter], inner_product[~mass_filter],
                     c=inner_product[~mass_filter], cmap="coolwarm",
                      vmin=vmin, vmax=vmax, s=s_grid)
    ax_.set_ylim([vmin*1.1, vmax*1.1])

    ax[1, 3].axhline(0, color="lightgray")
    pp = fig.colorbar(pcm, ax=ax_, orientation="vertical")
    sns.despine()

    ax[1, 3].set_xlabel("pseudotime")
    ax[1, 3].set_ylabel("inner product score")


def plot_scatter_with_anndata(adata, obsm_key, cluster_column_name, ax, args={}):

    embedding = adata.obsm[obsm_key]
    colors = _adata_to_color_dict(adata=adata, cluster_use=cluster_column_name)

    for cluster, color in colors.items():
        idx = np.where(adata.obs[cluster_column_name] == cluster)[0]
        ax.scatter(embedding[idx, 0], embedding[idx, 1], c=color, label=cluster, **args)



def visualize_developmental_analysis_ver2(self, scale_for_pseudotime=30, scale_for_simulated=30, s=10, s_grid=30, vmin=-1, vmax=1):

    whole_embedding = self.oracle.embedding
    original_embedding=self.oracle_dev.embedding
    original_value=self.oracle_dev.pseudotime
    mass_filter=self.oracle_dev.mass_filter
    grid=self.oracle_dev.flow_grid
    value_on_grid=self.oracle_dev.new_pseudotime
    gradient_pseudotime=self.oracle_dev.gradient
    gradient_simulated=self.oracle_dev.flow
    inner_product=self.oracle_dev.inner_product

    cluster = self.oracle_dev.cluster_loaded
    cluster_column_name = self.oracle_dev.cluster_column_name_loaded

    inner_product_stats = self.oracle_dev.inner_product_stats
    inner_product_stats_grouped = self.oracle_dev.inner_product_stats_grouped


    fig, ax = plt.subplots(2, 4, figsize=[20,10])

    if cluster == "whole":
        plot_scatter_with_anndata(adata=self.oracle.adata, obsm_key=self.obsm_key,
                                  cluster_column_name=cluster_column_name,
                                  ax=ax[0, 0], args={"s": s})
        ax[0, 0].set_title(f"Cluster of interest: all clusters")
        ax[0, 0].axis("off")
    else:
        cluster_color = _adata_to_color_dict(self.oracle.adata, cluster_column_name)[cluster]

        ax[0, 0].scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
        ax[0, 0].scatter(original_embedding[:, 0], original_embedding[:, 1], c=cluster_color, s=s)
        ax[0, 0].set_title(f"Cluster of interest: {cluster}")
        ax[0, 0].axis("off")

    ax[0, 1].scatter(original_embedding[:, 0], original_embedding[:, 1], c=original_value, cmap="rainbow", s=s)
    ax[0, 1].set_title("Pseudotime")
    ax[0, 1].axis("off")



    ax_ = ax[0, 2]
    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_pseudotime[~mass_filter, 0], gradient_pseudotime[~mass_filter, 1],
                    scale=scale_for_pseudotime)
    ax_.set_title("Gradient of pseudotime \n(=Development flow)")
    ax_.axis("off")

    ax_ = ax[0, 3]
    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_simulated[~mass_filter, 0], gradient_simulated[~mass_filter, 1],
                    scale=scale_for_simulated, zorder=20000)
    ax_.set_title("Perturb simulation")
    ax_.axis("off")


    ax[1, 0].scatter(grid[~mass_filter, 0], grid[~mass_filter, 1], c=inner_product[~mass_filter], cmap="coolwarm", s=s_grid, vmin=vmin, vmax=vmax)
    ax[1, 0].set_title("Inner product of \n Perturb simulation * Development flow")
    ax[1, 0].axis("off")


    ax[1, 1].scatter(grid[~mass_filter, 0], grid[~mass_filter, 1], c=inner_product[~mass_filter], cmap="coolwarm", s=s_grid, vmin=vmin, vmax=vmax)
    ax[1, 1].quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_simulated[~mass_filter, 0], gradient_simulated[~mass_filter, 1],
                    scale=scale_for_simulated, zorder=20000)
    ax[1, 1].axis("off")
    ax[1, 1].set_title("Inner product of \n Perturb simulation * Development flow \n + Perturb simulation")

    ax_ = ax[1, 2]
    pcm = ax_.scatter(value_on_grid[~mass_filter], inner_product[~mass_filter],
                     c=inner_product[~mass_filter], cmap="coolwarm",
                      vmin=vmin, vmax=vmax, s=s_grid)
    ax_.set_ylim([vmin*1.1, vmax*1.1])

    ax_.axhline(0, color="lightgray")
    pp = fig.colorbar(pcm, ax=ax_, orientation="vertical")
    sns.despine()

    ax_.set_xlabel("pseudotime")
    ax_.set_ylabel("inner product score")


    ax_ = ax[1, 3]
    sns.boxplot(data=inner_product_stats, x="pseudotime_id", y="score", color="white", ax=ax_)
    ax_.set_xlabel("Digitized_pseudotime")
    ax_.set_ylabel("inner product score")
    ax_.axhline(0, color="gray")
    ax_.set_ylim([vmin*1.1, vmax*1.1])
    ax_.tick_params(
                labelleft=False)




def visualize_developmental_analysis_ver101(self, scale_for_pseudotime=30, scale_for_simulated=30, s=10, s_grid=30, vmin=-1, vmax=1):

    whole_embedding = self.oracle.embedding
    original_embedding=self.oracle_dev.embedding
    original_value=self.oracle_dev.pseudotime
    mass_filter=self.oracle_dev.mass_filter
    grid=self.oracle_dev.flow_grid
    value_on_grid=self.oracle_dev.new_pseudotime
    gradient_pseudotime=self.oracle_dev.gradient
    gradient_simulated=self.oracle_dev.flow
    inner_product=self.oracle_dev.inner_product


    cluster = self.oracle_dev.cluster_loaded
    #if cluster == "True":
    #    cluster = True

    cluster_column_name = self.oracle_dev.cluster_column_name_loaded


    fig, ax = plt.subplots(2, 4, figsize=[20,10])

    ax_ = ax[0, 0]

    if cluster == "whole":
        plot_scatter_with_anndata(adata=self.oracle.adata, obsm_key=self.obsm_key,
                                  cluster_column_name=cluster_column_name,
                                  ax=ax[0, 0], args={"s": s})
        ax_.set_title(f"Cluster of interest: all clusters")

    else:
        cluster_color = _adata_to_color_dict(self.oracle.adata, cluster_column_name)[cluster]


        ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
        ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=cluster_color, s=s)
        if cluster == "True" :
            ax_.set_title(f"Cells of interest: \n{cluster_column_name}")

    ax_.axis("off")

    ##
    ax_ = ax[0, 1]
    ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
    ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=original_value, cmap="rainbow", s=s)
    ax_.set_title("Pseudotime")
    ax_.axis("off")

    ####
    ax_ = ax[0, 2]
    ax_.scatter(grid[mass_filter, 0], grid[mass_filter, 1], s=0)
    ax_.scatter(grid[~mass_filter, 0], grid[~mass_filter, 1], c=value_on_grid[~mass_filter], cmap="rainbow", s=s_grid)
    ax_.set_title("Pseudotime on grid")
    ax_.axis("off")


    ###
    ax_ = ax[0, 3]
    ax_.scatter(grid[mass_filter, 0], grid[mass_filter, 1], s=0)
    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_pseudotime[~mass_filter, 0], gradient_pseudotime[~mass_filter, 1],
                    scale=scale_for_pseudotime)
    ax_.set_title("Gradient of pseudotime \n(=Development flow)")
    ax_.axis("off")


    ####
    ax_ = ax[1, 0]
    ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
    ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=original_value, cmap="rainbow", s=s)
    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_pseudotime[~mass_filter, 0], gradient_pseudotime[~mass_filter, 1],
                    scale=scale_for_pseudotime)
    ax_.set_title("Pseudotime + \nDevelopment flow")
    ax_.axis("off")

    ####
    ax_ = ax[1, 1]
    ax_.scatter(grid[mass_filter, 0], grid[mass_filter, 1], s=0)
    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_simulated[~mass_filter, 0], gradient_simulated[~mass_filter, 1],
                    scale=scale_for_simulated, zorder=20000)
    ax_.set_title("Perturb simulation")
    ax_.axis("off")

    ax_ = ax[1, 2]
    ax_.scatter(grid[mass_filter, 0], grid[mass_filter, 1], s=0)
    ax_.scatter(grid[~mass_filter, 0], grid[~mass_filter, 1], c=inner_product[~mass_filter],
                cmap="coolwarm", s=s_grid, vmin=vmin, vmax=vmax)

    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_simulated[~mass_filter, 0], gradient_simulated[~mass_filter, 1],
                    scale=scale_for_simulated, zorder=20000)
    ax_.axis("off")
    ax_.set_title("Inner product of \n Perturb simulation * Development flow \n + Perturb simulation")

    #####
    ax_ = ax[1, 3]
    pcm = ax_.scatter(value_on_grid[~mass_filter], inner_product[~mass_filter],
                     c=inner_product[~mass_filter], cmap="coolwarm",
                      vmin=vmin, vmax=vmax, s=s_grid)

    ax_.set_ylim([vmin*1.1, vmax*1.1])
    ax_.axhline(0, color="lightgray")
    pp = fig.colorbar(pcm, ax=ax_, orientation="vertical")
    sns.despine()
    ax_.set_xlabel("pseudotime")
    ax_.set_ylabel("inner product score")



def visualize_developmental_analysis_ver201(self, scale_for_pseudotime=30, scale_for_simulated=30, s=10, s_grid=30, vmin=-1, vmax=1):

    whole_embedding = self.oracle.embedding
    original_embedding=self.oracle_dev.embedding
    original_value=self.oracle_dev.pseudotime
    mass_filter=self.oracle_dev.mass_filter
    grid=self.oracle_dev.flow_grid
    value_on_grid=self.oracle_dev.new_pseudotime
    gradient_pseudotime=self.oracle_dev.gradient
    gradient_simulated=self.oracle_dev.flow
    inner_product=self.oracle_dev.inner_product

    inner_product_stats = self.oracle_dev.inner_product_stats
    inner_product_stats_grouped = self.oracle_dev.inner_product_stats_grouped


    cluster = self.oracle_dev.cluster_loaded
    #if cluster == "True":
    #    cluster = True

    cluster_column_name = self.oracle_dev.cluster_column_name_loaded


    fig, ax = plt.subplots(2, 4, figsize=[20,10])

    ax_ = ax[0, 0]

    if cluster == "whole":
        plot_scatter_with_anndata(adata=self.oracle.adata, obsm_key=self.obsm_key,
                                  cluster_column_name=cluster_column_name,
                                  ax=ax_, args={"s": s})
        ax_.set_title(f"Cluster of interest: all clusters")

    else:
        cluster_color = _adata_to_color_dict(self.oracle.adata, cluster_column_name)[cluster]

        ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
        ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=cluster_color, s=s)
        if cluster == True :
            ax_.set_title(f"Cells of interest: \n{cluster_column_name}")

    ax_.axis("off")

    ##
    ax_ = ax[0, 1]
    ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
    ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=original_value, cmap="rainbow", s=s)
    ax_.set_title("Pseudotime")
    ax_.axis("off")

    ####

    ###
    ax_ = ax[0, 2]
    ax_.scatter(grid[mass_filter, 0], grid[mass_filter, 1], s=0)
    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_pseudotime[~mass_filter, 0], gradient_pseudotime[~mass_filter, 1],
                    scale=scale_for_pseudotime)
    ax_.set_title("Gradient of pseudotime \n(=Development flow)")
    ax_.axis("off")

    #####

    ax_ = ax[0, 3]
    ax_.scatter(grid[mass_filter, 0], grid[mass_filter, 1], s=0)
    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_simulated[~mass_filter, 0], gradient_simulated[~mass_filter, 1],
                    scale=scale_for_simulated, zorder=20000)
    ax_.set_title("Perturb simulation")
    ax_.axis("off")


    ####
    ax_ = ax[1, 0]
    ax_.scatter(grid[mass_filter, 0], grid[mass_filter, 1], s=0)
    ax_.scatter(grid[~mass_filter, 0], grid[~mass_filter, 1], c=inner_product[~mass_filter],
                cmap="coolwarm", s=s_grid, vmin=vmin, vmax=vmax)

    ax_.axis("off")
    ax_.set_title("Inner product of \n Perturb simulation * Development flow")


    ax_ = ax[1, 1]
    ax_.scatter(grid[mass_filter, 0], grid[mass_filter, 1], s=0)
    ax_.scatter(grid[~mass_filter, 0], grid[~mass_filter, 1], c=inner_product[~mass_filter],
                cmap="coolwarm", s=s_grid, vmin=vmin, vmax=vmax)

    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_simulated[~mass_filter, 0], gradient_simulated[~mass_filter, 1],
                    scale=scale_for_simulated, zorder=20000)
    ax_.axis("off")
    ax_.set_title("Inner product of \n Perturb simulation * Development flow \n + Perturb simulation")


    #####
    ax_ = ax[1, 2]
    pcm = ax_.scatter(value_on_grid[~mass_filter], inner_product[~mass_filter],
                     c=inner_product[~mass_filter], cmap="coolwarm",
                      vmin=vmin, vmax=vmax, s=s_grid)

    ax_.set_ylim([vmin*1.1, vmax*1.1])
    ax_.axhline(0, color="lightgray")
    pp = fig.colorbar(pcm, ax=ax_, orientation="vertical")
    sns.despine()
    ax_.set_xlabel("pseudotime")
    ax_.set_ylabel("inner product score")

    ax_ = ax[1, 3]
    sns.boxplot(data=inner_product_stats, x="pseudotime_id", y="score", color="white", ax=ax_)
    ax_.set_xlabel("Digitized_pseudotime")
    ax_.set_ylabel("inner product score")
    ax_.axhline(0, color="gray")
    ax_.set_ylim([vmin*1.1, vmax*1.1])
    ax_.tick_params(
                labelleft=False)



#####
def _get_ix_for_a_cluster(oracle, cluster_column_name, cluster):
    ix = np.arange(oracle.adata.shape[0])[oracle.adata.obs[cluster_column_name] == cluster]
    return ix

def _plot_quiver_for_a_cluster(oracle, cluster_column_name, cluster, quiver_scale, ax, color=None, plot_whole_cells=True, args={}):

    if cluster == "whole":
        ix_choice = ix = np.arange(oracle.adata.shape[0])
    else:
        ix_choice = _get_ix_for_a_cluster(oracle, cluster_column_name, cluster)


    if plot_whole_cells:

        ax.scatter(oracle.embedding[:, 0], oracle.embedding[:, 1],
                    c="lightgray", alpha=1, lw=0.3, rasterized=True,  **args)

    ax.scatter(oracle.embedding[ix_choice, 0], oracle.embedding[ix_choice, 1],
                c="lightgray", alpha=0.2, edgecolor=(0,0,0,1), lw=0.3, rasterized=True, **args)



    if color is None:
        color=oracle.colorandum[ix_choice]

    quiver_kwargs=dict(headaxislength=7, headlength=11, headwidth=8,
                       linewidths=0.25, width=0.0045,edgecolors="k",
                       color=color, alpha=1)

    ax.quiver(oracle.embedding[ix_choice, 0], oracle.embedding[ix_choice, 1],
               oracle.delta_embedding[ix_choice, 0],
               oracle.delta_embedding[ix_choice, 1],
               scale=quiver_scale, **quiver_kwargs)

    plt.axis("off")




def visualize_developmental_analysis_ver301(self, scale_for_pseudotime=30, scale_for_simulated=30, quiver_scale=30, s=10, s_grid=30, vmin=-1, vmax=1):

    whole_embedding = self.oracle.embedding
    original_embedding=self.oracle_dev.embedding
    original_value=self.oracle_dev.pseudotime
    mass_filter=self.oracle_dev.mass_filter
    grid=self.oracle_dev.flow_grid
    value_on_grid=self.oracle_dev.new_pseudotime
    gradient_pseudotime=self.oracle_dev.gradient
    gradient_simulated=self.oracle_dev.flow
    inner_product=self.oracle_dev.inner_product

    inner_product_stats = self.oracle_dev.inner_product_stats
    inner_product_stats_grouped = self.oracle_dev.inner_product_stats_grouped

    alpha = 1

    cluster = self.oracle_dev.cluster_loaded
    #if cluster == "True":
    #    cluster = True

    cluster_column_name = self.oracle_dev.cluster_column_name_loaded


    ''' fig, ax = plt.subplots(1, 4, figsize=[20,5])

    ax_ = ax[0]
    ax_.set_title(f"Clustering results")
    plot_scatter_with_anndata(adata=self.oracle.adata, obsm_key=self.obsm_key,,
                                  cluster_column_name=self.oracle.cluster_column_name,
                                  ax=ax_, args={"s": s})
    ax_.axis("off")'''


    fig, ax = plt.subplots(1, 4, figsize=[20,5])

    ax_ = ax[0]
    if cluster == "whole":
        plot_scatter_with_anndata(adata=self.oracle.adata, obsm_key=self.obsm_key,
                                  cluster_column_name=cluster_column_name,
                                  ax=ax_, args={"s": s})
        ax_.set_title(f"Cluster of interest: all clusters")

    else:
        cluster_color = _adata_to_color_dict(self.oracle.adata, cluster_column_name)[cluster]

        ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
        ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=cluster_color, s=s)
        if cluster == "True" :
            ax_.set_title(f"Cells of interest: \n{cluster_column_name}")
    ax_.axis("off")

    ##
    ax_ = ax[1]
    ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
    ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=original_value, cmap="rainbow", s=s)
    ax_.set_title("Pseudotime")
    ax_.axis("off")

    ####
    ###
    ax_ = ax[2]
    ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
    #ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=original_value, cmap="rainbow", s=s)
    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_pseudotime[~mass_filter, 0], gradient_pseudotime[~mass_filter, 1],
                    scale=scale_for_pseudotime)
    ax_.set_title("Gradient of pseudotime \n(=Development flow)")
    ax_.axis("off")


    ax_ = ax[3]
    if cluster == "whole":
        plot_scatter_with_anndata(adata=self.oracle.adata, obsm_key=self.obsm_key,
                                  cluster_column_name=cluster_column_name,
                                  ax=ax_, args={"s": s})
    else:
        ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
        ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=cluster_color, s=s, alpha=alpha)
    #ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=original_value, cmap="rainbow", s=s)
    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_pseudotime[~mass_filter, 0], gradient_pseudotime[~mass_filter, 1],
                    scale=scale_for_pseudotime)
    ax_.set_title("Gradient of pseudotime \n(=Development flow)")
    ax_.axis("off")

    plt.show()



    fig, ax = plt.subplots(1, 4, figsize=[20,5])


    #####
    ax_ = ax[0]
    ax_.set_title(f"Perturb simulation \n color: {cluster_column_name}")
    if cluster == "whole":
        _plot_quiver_for_a_cluster(oracle=self.oracle,
                                   cluster_column_name=cluster_column_name,
                                   color=None,
                                   cluster=cluster, quiver_scale=30, ax=ax_, args={"s": s})
    else:
        _plot_quiver_for_a_cluster(oracle=self.oracle,
                                   cluster_column_name=cluster_column_name,
                                   color=cluster_color,
                                   cluster=cluster, quiver_scale=30, ax=ax_, args={"s": s})
    ax_.axis("off")

    ax_ = ax[1]
    ax_.set_title("Perturb simulation \n color: cluster")
    _plot_quiver_for_a_cluster(oracle=self.oracle,
                               cluster_column_name=cluster_column_name,
                               cluster=cluster, quiver_scale=30, ax=ax_, args={"s": s})
    ax_.axis("off")

    ax_ = ax[2]
    ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_simulated[~mass_filter, 0], gradient_simulated[~mass_filter, 1],
                    scale=scale_for_simulated, zorder=20000)
    ax_.set_title("Perturb simulation result on grid")
    ax_.axis("off")


    ax_ = ax[3]
    if cluster == "whole":
        plot_scatter_with_anndata(adata=self.oracle.adata, obsm_key=self.obsm_key,
                                  cluster_column_name=cluster_column_name,
                                  ax=ax_, args={"s": s})
    else:
        ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
        ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=cluster_color, s=s, alpha=alpha)
    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_simulated[~mass_filter, 0], gradient_simulated[~mass_filter, 1],
                    scale=scale_for_simulated, zorder=20000)
    ax_.set_title("Perturb simulation result on grid")
    ax_.axis("off")

    plt.show()

    fig, ax = plt.subplots(1, 4, figsize=[20,5])

    ####
    ax_ = ax[0]
    ax_.scatter(grid[mass_filter, 0], grid[mass_filter, 1], s=0)
    ax_.scatter(grid[~mass_filter, 0], grid[~mass_filter, 1], c=inner_product[~mass_filter],
                cmap="coolwarm", s=s_grid, vmin=vmin, vmax=vmax)

    ax_.axis("off")
    ax_.set_title("Inner product of \n Perturb simulation * Development flow")


    ax_ = ax[1]
    ax_.scatter(grid[mass_filter, 0], grid[mass_filter, 1], s=0)
    ax_.scatter(grid[~mass_filter, 0], grid[~mass_filter, 1], c=inner_product[~mass_filter],
                cmap="coolwarm", s=s_grid, vmin=vmin, vmax=vmax)

    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_simulated[~mass_filter, 0], gradient_simulated[~mass_filter, 1],
                    scale=scale_for_simulated, zorder=20000)
    ax_.axis("off")
    ax_.set_title("Inner product of \n Perturb simulation * Development flow \n + Perturb simulation")


    #####
    ax_ = ax[2]
    pcm = ax_.scatter(value_on_grid[~mass_filter], inner_product[~mass_filter],
                     c=inner_product[~mass_filter], cmap="coolwarm",
                      vmin=vmin, vmax=vmax, s=s_grid)

    ax_.set_ylim([vmin*1.1, vmax*1.1])
    ax_.axhline(0, color="lightgray")
    pp = fig.colorbar(pcm, ax=ax_, orientation="vertical")
    sns.despine()
    ax_.set_xlabel("pseudotime")
    ax_.set_ylabel("inner product score")

    ax_ = ax[3]
    sns.boxplot(data=inner_product_stats, x="pseudotime_id", y="score", color="white", ax=ax_)
    ax_.set_xlabel("Digitized_pseudotime")
    ax_.set_ylabel("inner product score")
    ax_.axhline(0, color="gray")
    ax_.set_ylim([vmin*1.1, vmax*1.1])
    ax_.tick_params(
                labelleft=False)
    plt.show()



def visualize_developmental_analysis_ver401(self, scale_for_pseudotime=30, scale_for_simulated=30, quiver_scale=30, s=10, s_grid=30, vmin=-1, vmax=1):

    whole_embedding = self.oracle.embedding
    original_embedding=self.oracle_dev.embedding
    original_value=self.oracle_dev.pseudotime
    mass_filter=self.oracle_dev.mass_filter
    grid=self.oracle_dev.flow_grid
    value_on_grid=self.oracle_dev.new_pseudotime
    gradient_pseudotime=self.oracle_dev.gradient
    gradient_simulated=self.oracle_dev.flow
    inner_product=self.oracle_dev.inner_product

    inner_product_stats = self.oracle_dev.inner_product_stats
    inner_product_stats_grouped = self.oracle_dev.inner_product_stats_grouped

    alpha = 1

    cluster = self.oracle_dev.cluster_loaded
    #if cluster == "True":
    #    cluster = True

    cluster_column_name = self.oracle_dev.cluster_column_name_loaded


    ''' fig, ax = plt.subplots(1, 4, figsize=[20,5])

    ax_ = ax[0]
    ax_.set_title(f"Clustering results")
    plot_scatter_with_anndata(adata=self.oracle.adata, obsm_key=self.obsm_key,,
                                  cluster_column_name=self.oracle.cluster_column_name,
                                  ax=ax_, args={"s": s})
    ax_.axis("off")'''


    fig, ax = plt.subplots(1, 4, figsize=[20,5])

    ax_ = ax[0]
    if cluster == "whole":
        plot_scatter_with_anndata(adata=self.oracle.adata, obsm_key=self.obsm_key,
                                  cluster_column_name=cluster_column_name,
                                  ax=ax_, args={"s": s})
        ax_.set_title(f"Cluster of interest: all clusters")

    else:
        cluster_color = _adata_to_color_dict(self.oracle.adata, cluster_column_name)[cluster]

        ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
        ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=cluster_color, s=s)
        if cluster == "True" :
            ax_.set_title(f"Cells of interest: \n{cluster_column_name}")
    ax_.axis("off")

    ##
    ax_ = ax[1]
    ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
    ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=original_value, cmap="rainbow", s=s)
    ax_.set_title("Pseudotime")
    ax_.axis("off")

    ####
    ###
    ax_ = ax[2]
    ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
    #ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=original_value, cmap="rainbow", s=s)
    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_pseudotime[~mass_filter, 0], gradient_pseudotime[~mass_filter, 1],
                    scale=scale_for_pseudotime)
    ax_.set_title("Gradient of pseudotime \n(=Development flow)")
    ax_.axis("off")


    ax_ = ax[3]
    if cluster == "whole":
        plot_scatter_with_anndata(adata=self.oracle.adata, obsm_key=self.obsm_key,
                                  cluster_column_name=cluster_column_name,
                                  ax=ax_, args={"s": s})
    else:
        ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
        ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=cluster_color, s=s, alpha=alpha)
    #ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=original_value, cmap="rainbow", s=s)
    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_pseudotime[~mass_filter, 0], gradient_pseudotime[~mass_filter, 1],
                    scale=scale_for_pseudotime)
    ax_.set_title("Gradient of pseudotime \n(=Development flow)")
    ax_.axis("off")

    plt.show()



    fig, ax = plt.subplots(1, 4, figsize=[20,5])


    #####
    ax_ = ax[0]
    ax_.set_title(f"Perturb simulation \n color: {cluster_column_name}")
    if cluster == "whole":
        _plot_quiver_for_a_cluster(oracle=self.oracle,
                                   cluster_column_name=cluster_column_name,
                                   color=None,
                                   cluster=cluster, quiver_scale=30, ax=ax_, args={"s": s})
    else:
        _plot_quiver_for_a_cluster(oracle=self.oracle,
                                   cluster_column_name=cluster_column_name,
                                   color=cluster_color,
                                   cluster=cluster, quiver_scale=30, ax=ax_, args={"s": s})
    ax_.axis("off")

    ax_ = ax[1]
    ax_.set_title("Perturb simulation \n color: cluster")
    _plot_quiver_for_a_cluster(oracle=self.oracle,
                               cluster_column_name=cluster_column_name,
                               cluster=cluster, quiver_scale=30, ax=ax_, args={"s": s})
    ax_.axis("off")

    ax_ = ax[2]
    ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_simulated[~mass_filter, 0], gradient_simulated[~mass_filter, 1],
                    scale=scale_for_simulated, zorder=20000)
    ax_.set_title("Perturb simulation result on grid")
    ax_.axis("off")


    ax_ = ax[3]
    if cluster == "whole":
        plot_scatter_with_anndata(adata=self.oracle.adata, obsm_key=self.obsm_key,
                                  cluster_column_name=cluster_column_name,
                                  ax=ax_, args={"s": s})
    else:
        ax_.scatter(whole_embedding[:, 0], whole_embedding[:, 1], c="lightgray", s=s)
        ax_.scatter(original_embedding[:, 0], original_embedding[:, 1], c=cluster_color, s=s, alpha=alpha)
    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_simulated[~mass_filter, 0], gradient_simulated[~mass_filter, 1],
                    scale=scale_for_simulated, zorder=20000)
    ax_.set_title("Perturb simulation result on grid")
    ax_.axis("off")

    plt.show()

    fig, ax = plt.subplots(1, 4, figsize=[20,5])

    ####
    ax_ = ax[0]
    ax_.scatter(grid[mass_filter, 0], grid[mass_filter, 1], s=0)
    ax_.scatter(grid[~mass_filter, 0], grid[~mass_filter, 1], c=inner_product[~mass_filter],
                cmap="coolwarm", s=s_grid, vmin=vmin, vmax=vmax)

    ax_.axis("off")
    ax_.set_title("Inner product of \n Perturb simulation * Development flow")


    ax_ = ax[1]
    ax_.scatter(grid[mass_filter, 0], grid[mass_filter, 1], s=0)
    ax_.scatter(grid[~mass_filter, 0], grid[~mass_filter, 1], c=inner_product[~mass_filter],
                cmap="coolwarm", s=s_grid, vmin=vmin, vmax=vmax)

    ax_.quiver(grid[~mass_filter, 0], grid[~mass_filter, 1],
                    gradient_simulated[~mass_filter, 0], gradient_simulated[~mass_filter, 1],
                    scale=scale_for_simulated, zorder=20000)
    ax_.axis("off")
    ax_.set_title("Inner product of \n Perturb simulation * Development flow \n + Perturb simulation")


    #####
    ax_ = ax[2]
    pcm = ax_.scatter(value_on_grid[~mass_filter], inner_product[~mass_filter],
                     c=inner_product[~mass_filter], cmap="coolwarm",
                      vmin=vmin, vmax=vmax, s=s_grid)

    ax_.set_ylim([vmin*1.1, vmax*1.1])
    ax_.axhline(0, color="lightgray")
    #pp = fig.colorbar(pcm, ax=ax_, orientation="vertical")
    sns.despine()
    ax_.set_xlabel("pseudotime")
    ax_.set_ylabel("inner product score")

    ax_ = ax[3]
    sns.boxplot(data=inner_product_stats, x="pseudotime_id", y="score", color="white", ax=ax_)
    ax_.set_xlabel("Digitized_pseudotime")
    ax_.set_ylabel("inner product score")
    ax_.axhline(0, color="gray")
    ax_.set_ylim([vmin*1.1, vmax*1.1])
    ax_.tick_params(
                labelleft=False)
    plt.show()



    fig, ax = plt.subplots(1, 4, figsize=[20,5])

    stage_grid = self.oracle_dev.stage_grid
    colors_dict = _adata_to_color_dict(adata=self.oracle.adata, cluster_use="Stage")
    stage_order = return_order(stage_list=self.oracle_dev.inner_product_stats.stage)


    ####
    ax_ = ax[0]
    ax_.scatter(grid[mass_filter, 0], grid[mass_filter, 1], s=0)
    plot_grid_with_categprocal_color(grid=grid, mass_filter=mass_filter, color_array=stage_grid,
                                 colors_dict=colors_dict,
                                 ax=ax_, args={})

    ax_.axis("off")
    ax_.set_title("Stage (hpf) on grid")


    ax_ = ax[1]
    plot_legend(labels=stage_order, palette=colors_dict, ax_=ax_)

    #####
    ax_ = ax[2]
    sns.violinplot(data=self.oracle_dev.inner_product_stats,
                   x="pseudotime", y="stage",
                   palette=colors_dict, order=stage_order[::-1],
                   ax=ax_)
    sns.despine()
    #ax_.set_xlabel("pseudotime")
    ax_.set_ylabel("stage (hpf)")

    ax_ = ax[3]
    plot_stackedvar(pd.crosstab(self.oracle_dev.inner_product_stats['stage'],
                                self.oracle_dev.inner_product_stats['pseudotime_id']).loc[stage_order],
                                ax=ax_, palette=colors_dict)
    ax_.set_xlabel("Digitized_pseudotime")
    ax_.set_ylabel("Grid point count")
    sns.despine()
    plt.show()

def plot_grid_with_categprocal_color(grid, mass_filter, color_array, colors_dict, ax, args={}):
    x = grid[~mass_filter, 0]
    y = grid[~mass_filter, 1]
    color_array_filtered = color_array[~mass_filter]

    for cluster, color in colors_dict.items():
        idx = np.where(color_array_filtered == cluster)[0]
        ax.scatter(x[idx], y[idx], c=color, label=cluster, **args)

def return_order(stage_list):
    stage_unique = np.unique(stage_list)
    hpf_unique = [float(i[:4]) for i in stage_unique]
    stage_order = list(stage_unique[np.argsort(hpf_unique)])
    return stage_order

def plot_legend(labels, palette, ax_):

    for i, label in enumerate(labels):
        ax_.scatter([0], [i], s=100, c=palette[label])
        ax_.text(1, i-len(labels)*0.015, s=label)
    ax_.set_ylim([-1, len(labels)])
    ax_.set_xlim([-1, 10])
    ax_.axis("off")

def plot_stackedvar(df, ax, palette=None):

    bottom_feats=[]
    if palette is None:
        for i, j in enumerate(df.index.values):
            if i==0:
                ax.bar(df.columns.values, df.loc[j].values, edgecolor='white', label=j)
            else:
                ax.bar(df.columns.values, df.loc[j].values, label=j,
                        bottom=df.loc[bottom_feats].sum(axis=0).values,
                        edgecolor='white')
            bottom_feats.append(j)
    else:
        for i, j in enumerate(df.index.values):
            if i==0:
                ax.bar(df.columns.values, df.loc[j].values,
                    edgecolor='white', color=palette[j], label=j)
            else:
                ax.bar(df.columns.values, df.loc[j].values, label=j, color=palette[j],
                        bottom=df.loc[bottom_feats].sum(axis=0).values,
                        edgecolor='white')
            bottom_feats.append(j)
    #plt.legend()
    ax.set_xticks(df.columns)



import os
import sys

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns


import celloracle as co
print("co version: ", co.__version__)

plt.rcParams['figure.figsize'] = [6, 4.5]
plt.rcParams["savefig.dpi"] = 300

def main():
    test_tutorial_network_analysis()

def test_tutorial_network_analysis():
    use_reduced_data = True

    # Start analysis
    import time
    start = time.time()

    save_folder = "figures"
    os.makedirs(save_folder, exist_ok=True)


    # 1. Load data
    print("\n1. Load data")
    adata = co.data.load_Paul2015_data()

    # Testing with whole data will take long time. We can test function with subsampled data.
    if use_reduced_data:
        cells = adata.obs.index[adata.obs.louvain_annot.isin(["MEP_0", "GMPl_0"])]
        adata = adata[cells, :]

    # (Option) Down sampling
    print(f"Cell number is :{adata.shape[0]}")
    print(f"Gene number is :{adata.shape[1]}")
    # Random downsampling into 30K cells if the anndata object include more than 30 K cells.
    n_cells_downsample = 30000
    if adata.shape[0] > n_cells_downsample:
        # Let's dowmsample into 30K cells
        sc.pp.subsample(adata, n_obs=n_cells_downsample, random_state=123)
    print(f"Cell number is :{adata.shape[0]}")
    # Load TF info which was made from mouse cell atlas dataset.
    base_GRN = co.data.load_mouse_scATAC_atlas_base_GRN()
    # Check data
    base_GRN.head()

    # 2. Make Oracle object
    print("\n2. Make Oracle object")
    # Instantiate Oracle object
    oracle = co.Oracle()
    # Check data in anndata
    print("Metadata columns :", list(adata.obs.columns))
    print("Dimensional reduction: ", list(adata.obsm.keys()))
    # In this notebook, we use the unscaled mRNA count for the nput of Oracle object.
    adata.X = adata.layers["raw_count"].copy()
    # Instantiate Oracle object.
    oracle.import_anndata_as_raw_count(adata=adata,
                                       cluster_column_name="louvain_annot",
                                       embedding_name="X_draw_graph_fa")
    # You can load TF info dataframe with the following code.
    oracle.import_TF_data(TF_info_matrix=base_GRN)

    ## 2.3. (Optional) Add TF-target gene pair manually
    print("\n (Optional) Add TF-target gene pair manually")
    # Download TF-target gene pair data
    from celloracle.utility.data_download_from_web import download_demo_data
    for i in ["TF_data_in_Paul15.csv"]:
        download_demo_data(file=i)
    # Load the TF and target gene information from Paul et al. (2015).
    Paul_15_data = pd.read_csv("TF_data_in_Paul15.csv")
    Paul_15_data
    # Make dictionary: dictionary key is TF and dictionary value is list of target genes.
    TF_to_TG_dictionary = {}
    for TF, TGs in zip(Paul_15_data.TF, Paul_15_data.Target_genes):
        # convert target gene to list
        TG_list = TGs.replace(" ", "").split(",")
        # store target gene list in a dictionary
        TF_to_TG_dictionary[TF] = TG_list
    # We invert the dictionary above using a utility function in celloracle.
    TG_to_TF_dictionary = co.utility.inverse_dictionary(TF_to_TG_dictionary)
    # Add TF information
    oracle.addTFinfo_dictionary(TG_to_TF_dictionary)

    # 3. KNN imputation
    print("\n3. KNN imputation")
    # Perform PCA
    oracle.perform_PCA()
    # Select important PCs
    plt.plot(np.cumsum(oracle.pca.explained_variance_ratio_)[:100])
    n_comps = np.where(np.diff(np.diff(np.cumsum(oracle.pca.explained_variance_ratio_))>0.002))[0][0]
    plt.axvline(n_comps, c="k")
    plt.show()
    print(n_comps)
    n_comps = min(n_comps, 50)
    # KNN imputation
    n_cell = oracle.adata.shape[0]
    print(f"cell number is :{n_cell}")
    k = int(0.025*n_cell)
    print(f"Auto-selected k is :{k}")
    oracle.knn_imputation(n_pca_dims=n_comps, k=k, balanced=True, b_sight=k*8,
                          b_maxl=k*4, n_jobs=4)

    # 4. Save and Load.
    print("\n4. Save and Load.")
    # Save oracle object.
    oracle.to_hdf5("Paul_15_data.celloracle.oracle")
    # Load file.
    oracle = co.load_hdf5("Paul_15_data.celloracle.oracle")

    # 5. GRN calculation
    print("\n5. GRN calculation")
    # Check clustering data
    sc.pl.draw_graph(oracle.adata, color="louvain_annot")
    # Calculate GRN for each population in "louvain_annot" clustering unit.
    # This step may take some time.(~30 minutes)
    links = oracle.get_links(cluster_name_for_GRN_unit="louvain_annot",
                             alpha=10,
                             verbose_level=1)
    # (Optional) Export GRNs
    links.links_dict.keys()

    if use_reduced_data:
        pass
    else:
        # 5.3. (Optional) Change order
        # Show the contents of pallete
        links.palette
        # Change the order of pallete
        order = ['MEP_0', 'Mk_0', 'Ery_0',
                 'Ery_1', 'Ery_2', 'Ery_3', 'Ery_4', 'Ery_5',
                 'Ery_6', 'Ery_7', 'Ery_8', 'Ery_9',
                 'GMP_0', 'GMP_1', 'GMP_2', 'GMPl_0', 'GMPl_1',
                 'Mo_0', 'Mo_1', 'Mo_2',
                 'Gran_0', 'Gran_1', 'Gran_2', 'Gran_3']
        links.palette = links.palette.loc[order]
        links.palette
    # Save Links object.
    links.to_hdf5(file_path="links.celloracle.links")

    # 6. Network preprocessing
    print("\n6. Network preprocessing")
    links.filter_links(p=0.001, weight="coef_abs", threshold_number=2000)

    print("\n")
    plt.rcParams["figure.figsize"] = [9, 4.5]
    links.plot_degree_distributions(plot_model=True,
                                    save=f"{save_folder}/degree_distribution/",
                                    )
    plt.rcParams["figure.figsize"] = [6, 4.5]
    # Calculate network scores.
    links.get_network_score()
    links.merged_score.head()
    # Save Links object.
    links.to_hdf5(file_path="links.celloracle.links")
    # You can load files with the following command.
    links = co.load_hdf5(file_path="links.celloracle.links")

    # 7. Network analysis; Network score for each gene
    print("\n7. Network analysis; Network score for each gene")
    # Check cluster name
    links.cluster
    # Visualize top n-th genes with high scores.
    links.plot_scores_as_rank(cluster="MEP_0", n_gene=30,
                              save=f"{save_folder}/ranked_score")
    # Compare GRN score between two clusters
    links.plot_score_comparison_2D(value="eigenvector_centrality",
                                   cluster1="MEP_0", cluster2="GMPl_0",
                                   percentile=98,
                                   save=f"{save_folder}/score_comparison")
    # Compare GRN score between two clusters
    links.plot_score_comparison_2D(value="betweenness_centrality",
                                   cluster1="MEP_0", cluster2="GMPl_0",
                                   percentile=98,
                                   save=f"{save_folder}/score_comparison")
    # Compare GRN score between two clusters
    links.plot_score_comparison_2D(value="degree_centrality_all",
                                   cluster1="MEP_0", cluster2="GMPl_0",
                                   percentile=98,
                                   save=f"{save_folder}/score_comparison")
    # Visualize Gata2 network score dynamics
    links.plot_score_per_cluster(goi="Gata2",
                                 save=f"{save_folder}/network_score_per_gene/")
    # Visualize Cebpa network score dynamics
    links.plot_score_per_cluster(goi="Cebpa")
    #
    

    # 8. Network analysis; network score distribution
    print("\n8. Network analysis; network score distribution")
    plt.rcParams["figure.figsize"] = [6, 4.5]
    # Plot degree_centrality
    plt.subplots_adjust(left=0.15, bottom=0.3)
    plt.ylim([0,0.040])
    links.plot_score_discributions(values=["degree_centrality_all"],
                                   method="boxplot",
                                   save=f"{save_folder}",
                                  )
    # Plot eigenvector_centrality
    plt.subplots_adjust(left=0.15, bottom=0.3)
    plt.ylim([0, 0.28])
    links.plot_score_discributions(values=["eigenvector_centrality"],
                                   method="boxplot",
                                   save=f"{save_folder}")
    ## Distribution of netowrk entropy
    plt.subplots_adjust(left=0.15, bottom=0.3)
    links.plot_network_entropy_distributions(save=f"{save_folder}")

    # 6. Remove files
    print("Remove files")
    for i in ["TF_data_in_Paul15.csv", "Paul_15_data.celloracle.oracle",
              "links.celloracle.links"]:
        os.remove(i)
    os.system("rm -r figures")

    elapsed = time.time() - start

    print("Success")
    print(f"Total time: {elapsed/60:.3g} minutes")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
'''
This file contains custom functions for the analysis of ATAC-seq data.
Genomic activity information (peak of ATAC-seq) will be extracted first.
Then the peak DNA sequence will be subjected to TF motif scan.
Finally we will get list of TFs that potentially binds to a specific gene.

Codes were written by Kenji Kamimoto.


'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing
import warnings
import pandas as pd
import numpy as np
import scanpy as sc
import sys, os
from ..data import __path__ as parent_path
from ..data.config import CELLORACLE_DATA_DIR, WEB_PAR_DIR
from ..utility import load_hdf5
from ..utility.data_download_from_web import download_data_if_data_not_exist


def load_drosophila_promoter_base_GRN(version="dm6_CisBPv2_fpr2", force_download=False):
    """
    Load Base GRN made from promoter DNA sequence and motif scan.

    Args:

    Returns:
        pandas.dataframe: Base GRN as a matrix.
    """

    options = {"dm3_CisBPv2_fpr1": "promoter_base_GRN/dm3_TFinfo_dataframe_CisBPv2_fpr1_threshold_10_20210630.parquet",
               "dm3_CisBPv2_fpr2": "promoter_base_GRN/dm3_TFinfo_dataframe_CisBPv2_fpr2_threshold_10_20210630.parquet",
               "dm6_CisBPv2_fpr1": "promoter_base_GRN/dm6_TFinfo_dataframe_CisBPv2_fpr1_threshold_10_20210630.parquet",
               "dm6_CisBPv2_fpr2": "promoter_base_GRN/dm6_TFinfo_dataframe_CisBPv2_fpr2_threshold_10_20210630.parquet",
               }
    if version in options.keys():
        print(f"Loading prebuilt promoter base-GRN. Version: {version}")
        filename = options[version]
    else:
        print(f"Version error. {version} is not in the list.")
        print("Available option: ", list(options.keys()))

    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], filename)
    if (force_download == False) & os.path.isfile(path):
        pass
    else:
        path = os.path.join(CELLORACLE_DATA_DIR, filename)
        backup_url = os.path.join(WEB_PAR_DIR, filename)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)

    return pd.read_parquet(path)

def load_rat_promoter_base_GRN(version="rn6_gimmemotifsv5_fpr2", force_download=False):
    """
    Load Base GRN made from promoter DNA sequence and motif scan.

    Args:

    Returns:
        pandas.dataframe: Base GRN as a matrix.
    """

    options = {"rn4_gimmemotifsv5_fpr1": "promoter_base_GRN/rn4_TFinfo_dataframe_gimmemotifsv5_fpr1_threshold_10_20210630.parquet",
               "rn4_gimmemotifsv5_fpr2": "promoter_base_GRN/rn4_TFinfo_dataframe_gimmemotifsv5_fpr2_threshold_10_20210630.parquet",
               "rn5_gimmemotifsv5_fpr1": "promoter_base_GRN/rn5_TFinfo_dataframe_gimmemotifsv5_fpr1_threshold_10_20210630.parquet",
               "rn5_gimmemotifsv5_fpr2": "promoter_base_GRN/rn5_TFinfo_dataframe_gimmemotifsv5_fpr2_threshold_10_20210630.parquet",
               "rn6_gimmemotifsv5_fpr1": "promoter_base_GRN/rn6_TFinfo_dataframe_gimmemotifsv5_fpr1_threshold_10_20210630.parquet",
               "rn6_gimmemotifsv5_fpr2": "promoter_base_GRN/rn6_TFinfo_dataframe_gimmemotifsv5_fpr2_threshold_10_20210630.parquet",
               }
    if version in options.keys():
        print(f"Loading prebuilt promoter base-GRN. Version: {version}")
        filename = options[version]
    else:
        print(f"Version error. {version} is not in the list.")
        print("Available option: ", list(options.keys()))

    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], filename)
    if (force_download == False) & os.path.isfile(path):
        pass
    else:
        path = os.path.join(CELLORACLE_DATA_DIR, filename)
        backup_url = os.path.join(WEB_PAR_DIR, filename)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)

    return pd.read_parquet(path)


def load_mouse_promoter_base_GRN(version="mm10_gimmemotifsv5_fpr2", force_download=False):
    """
    Load Base GRN made from promoter DNA sequence and motif scan.

    Args:

    Returns:
        pandas.dataframe: Base GRN as a matrix.
    """

    options = {"mm9_gimmemotifsv5_fpr1": "promoter_base_GRN/mm9_TFinfo_dataframe_gimmemotifsv5_fpr1_threshold_10_20210630.parquet",
               "mm9_gimmemotifsv5_fpr2": "promoter_base_GRN/mm9_TFinfo_dataframe_gimmemotifsv5_fpr2_threshold_10_20210630.parquet",
               "mm10_gimmemotifsv5_fpr1": "promoter_base_GRN/mm10_TFinfo_dataframe_gimmemotifsv5_fpr1_threshold_10_20210630.parquet",
               "mm10_gimmemotifsv5_fpr2": "promoter_base_GRN/mm10_TFinfo_dataframe_gimmemotifsv5_fpr2_threshold_10_20210630.parquet",
               }
    if version in options.keys():
        print(f"Loading prebuilt promoter base-GRN. Version: {version}")
        filename = options[version]
    else:
        print(f"Version error. {version} is not in the list.")
        print("Available option: ", list(options.keys()))

    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], filename)
    if (force_download == False) & os.path.isfile(path):
        pass
    else:
        path = os.path.join(CELLORACLE_DATA_DIR, filename)
        backup_url = os.path.join(WEB_PAR_DIR, filename)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)

    return pd.read_parquet(path)

def load_human_promoter_base_GRN(version="hg19_gimmemotifsv5_fpr2", force_download=False):
    """
    Load Base GRN made from promoter DNA sequence and motif scan.

    Args:

    Returns:
        pandas.dataframe: Base GRN as a matrix.
    """

    options = {"hg19_gimmemotifsv5_fpr1": "promoter_base_GRN/hg19_TFinfo_dataframe_gimmemotifsv5_fpr1_threshold_10_20210630.parquet",
               "hg19_gimmemotifsv5_fpr2": "promoter_base_GRN/hg19_TFinfo_dataframe_gimmemotifsv5_fpr2_threshold_10_20210630.parquet",
               "hg38_gimmemotifsv5_fpr1": "promoter_base_GRN/hg38_TFinfo_dataframe_gimmemotifsv5_fpr1_threshold_10_20210630.parquet",
               "hg38_gimmemotifsv5_fpr2": "promoter_base_GRN/hg38_TFinfo_dataframe_gimmemotifsv5_fpr2_threshold_10_20210630.parquet",
               }
    if version in options.keys():
        print(f"Loading prebuilt promoter base-GRN. Version: {version}")
        filename = options[version]
    else:
        print(f"Version error. {version} is not in the list.")
        print("Available option: ", list(options.keys()))

    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], filename)
    if (force_download == False) & os.path.isfile(path):
        pass
    else:
        path = os.path.join(CELLORACLE_DATA_DIR, filename)
        backup_url = os.path.join(WEB_PAR_DIR, filename)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)

    return pd.read_parquet(path)


def load_chicken_promoter_base_GRN(version="galGal6_CisBPv2_fpr2", force_download=False):
    """
    Load Base GRN made from promoter DNA sequence and motif scan.

    Args:

    Returns:
        pandas.dataframe: Base GRN as a matrix.
    """

    options = {"galGal4_CisBPv2_fpr1": "promoter_base_GRN/galGal4_TFinfo_dataframe_CisBPv2_fpr1_threshold_10_20210630.parquet",
               "galGal4_CisBPv2_fpr2": "promoter_base_GRN/galGal4_TFinfo_dataframe_CisBPv2_fpr2_threshold_10_20210630.parquet",
               "galGal5_CisBPv2_fpr1": "promoter_base_GRN/galGal5_TFinfo_dataframe_CisBPv2_fpr1_threshold_10_20210630.parquet",
               "galGal5_CisBPv2_fpr2": "promoter_base_GRN/galGal5_TFinfo_dataframe_CisBPv2_fpr2_threshold_10_20210630.parquet",
               "galGal6_CisBPv2_fpr1": "promoter_base_GRN/galGal6_TFinfo_dataframe_CisBPv2_fpr1_threshold_10_20210630.parquet",
               "galGal6_CisBPv2_fpr2": "promoter_base_GRN/galGal6_TFinfo_dataframe_CisBPv2_fpr2_threshold_10_20210630.parquet",
               }
    if version in options.keys():
        print(f"Loading prebuilt promoter base-GRN. Version: {version}")
        filename = options[version]
    else:
        print(f"Version error. {version} is not in the list.")
        print("Available option: ", list(options.keys()))

    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], filename)
    if (force_download == False) & os.path.isfile(path):
        pass
    else:
        path = os.path.join(CELLORACLE_DATA_DIR, filename)
        backup_url = os.path.join(WEB_PAR_DIR, filename)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)

    return pd.read_parquet(path)


def load_zebrafish_promoter_base_GRN(version="danRer11_CisBPv2_fpr2", force_download=False):
    """
    Load Base GRN made from promoter DNA sequence and motif scan.

    Args:

    Returns:
        pandas.dataframe: Base GRN as a matrix.
    """

    options = {"danRer7_CisBPv2_fpr1": "promoter_base_GRN/danRer7_TFinfo_dataframe_CisBPv2_fpr1_threshold_10_20210630.parquet",
               "danRer7_CisBPv2_fpr2": "promoter_base_GRN/danRer7_TFinfo_dataframe_CisBPv2_fpr2_threshold_10_20210630.parquet",
               "danRer10_CisBPv2_fpr1": "promoter_base_GRN/danRer10_TFinfo_dataframe_CisBPv2_fpr1_threshold_10_20210630.parquet",
               "danRer10_CisBPv2_fpr2": "promoter_base_GRN/danRer10_TFinfo_dataframe_CisBPv2_fpr2_threshold_10_20210630.parquet",
               "danRer11_CisBPv2_fpr1": "promoter_base_GRN/danRer11_TFinfo_dataframe_CisBPv2_fpr1_threshold_10_20210630.parquet",
               "danRer11_CisBPv2_fpr2": "promoter_base_GRN/danRer11_TFinfo_dataframe_CisBPv2_fpr2_threshold_10_20210630.parquet",
               }
    if version in options.keys():
        print(f"Loading prebuilt promoter base-GRN. Version: {version}")
        filename = options[version]
    else:
        print(f"Version error. {version} is not in the list.")
        print("Available option: ", list(options.keys()))

    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], filename)
    if (force_download == False) & os.path.isfile(path):
        pass
    else:
        path = os.path.join(CELLORACLE_DATA_DIR, filename)
        backup_url = os.path.join(WEB_PAR_DIR, filename)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)

    return pd.read_parquet(path)

def load_xenopus_tropicalis_promoter_base_GRN(version="xenTro3_CisBPv2_fpr2", force_download=False):
    """
    Load Base GRN made from promoter DNA sequence and motif scan.

    Args:

    Returns:
        pandas.dataframe: Base GRN as a matrix.
    """
    warnings.warn("Promoter base GRN for Xenopus tropicalis has been updated since celloracle v0.10.14 duet to the change of default motif data base for Xenopus tropicalis. If you want to use previous version, please select old version by specifying version argument.",
                  UserWarning)

    options = {"xenTro2_CisBPv2_fpr1_v20210630": "promoter_base_GRN/xenTro2_TFinfo_dataframe_CisBPv2_Xenopus_tropicalis_and_Xenopus_laevis_fpr1_threshold_10_20210630.parquet",
               "xenTro2_CisBPv2_fpr2_v20210630": "promoter_base_GRN/xenTro2_TFinfo_dataframe_CisBPv2_Xenopus_tropicalis_and_Xenopus_laevis_fpr2_threshold_10_20210630.parquet",
               "xenTro3_CisBPv2_fpr1_v20210630": "promoter_base_GRN/xenTro3_TFinfo_dataframe_CisBPv2_Xenopus_tropicalis_and_Xenopus_laevis_fpr1_threshold_10_20210630.parquet",
               "xenTro3_CisBPv2_fpr2_v20210630": "promoter_base_GRN/xenTro3_TFinfo_dataframe_CisBPv2_Xenopus_tropicalis_and_Xenopus_laevis_fpr2_threshold_10_20210630.parquet",
               "xenTro2_CisBPv2_fpr1": "promoter_base_GRN/xenTro2_TFinfo_dataframe_CisBPv2_Xenopus_tropicalis_fpr1_threshold_10_20221231.parquet",
               "xenTro2_CisBPv2_fpr2": "promoter_base_GRN/xenTro2_TFinfo_dataframe_CisBPv2_Xenopus_tropicalis_fpr2_threshold_10_20221231.parquet",
               "xenTro3_CisBPv2_fpr1": "promoter_base_GRN/xenTro3_TFinfo_dataframe_CisBPv2_Xenopus_tropicalis_fpr1_threshold_10_20221231.parquet",
               "xenTro3_CisBPv2_fpr2": "promoter_base_GRN/xenTro3_TFinfo_dataframe_CisBPv2_Xenopus_tropicalis_fpr2_threshold_10_20221231.parquet",
               }

    if version in options.keys():
        print(f"Loading prebuilt promoter base-GRN. Version: {version}")
        filename = options[version]
    else:
        print(f"Version error. {version} is not in the list.")
        print("Available option: ", list(options.keys()))

    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], filename)
    if (force_download == False) & os.path.isfile(path):
        pass
    else:
        path = os.path.join(CELLORACLE_DATA_DIR, filename)
        backup_url = os.path.join(WEB_PAR_DIR, filename)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)

    return pd.read_parquet(path)

def load_xenopus_laevis_promoter_base_GRN(version="Xenopus_laevis_v10.1_CisBPv2_fpr2", force_download=False):
    """
    Load Base GRN made from promoter DNA sequence and motif scan.

    Args:

    Returns:
        pandas.dataframe: Base GRN as a matrix.
    """

    options = {"Xenopus_laevis_v10.1_CisBPv2_fpr2": "promoter_base_GRN/Xenopus_laevis_v10.1_TFinfo_dataframe_CisBPv2_Xenopus_laevis_fpr2_threshold_10_20221228.parquet",
               "Xenopus_laevis_v10.1_CisBPv2_fpr1": "promoter_base_GRN/Xenopus_laevis_v10.1_TFinfo_dataframe_CisBPv2_Xenopus_laevis_fpr1_threshold_10_20221228.parquet",

               }

    if version in options.keys():
        print(f"Loading prebuilt promoter base-GRN. Version: {version}")
        filename = options[version]
    else:
        print(f"Version error. {version} is not in the list.")
        print("Available option: ", list(options.keys()))

    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], filename)
    if (force_download == False) & os.path.isfile(path):
        pass
    else:
        path = os.path.join(CELLORACLE_DATA_DIR, filename)
        backup_url = os.path.join(WEB_PAR_DIR, filename)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)

    return pd.read_parquet(path)

def load_arabidopsis_promoter_base_GRN(version="TAIR10_CisBPv2_fpr2", force_download=False):
    """
    Load Base GRN made from promoter DNA sequence and motif scan.

    Args:

    Returns:
        pandas.dataframe: Base GRN as a matrix.
    """

    options = {"TAIR10_CisBPv2_fpr1": "promoter_base_GRN/TAIR10_TFinfo_dataframe_CisBPv2_fpr1_threshold_10_20210630.parquet",
               "TAIR10_CisBPv2_fpr2": "promoter_base_GRN/TAIR10_TFinfo_dataframe_CisBPv2_fpr2_threshold_10_20210630.parquet",
               }
    if version in options.keys():
        print(f"Loading prebuilt promoter base-GRN. Version: {version}")
        filename = options[version]
    else:
        print(f"Version error. {version} is not in the list.")
        print("Available option: ", list(options.keys()))

    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], filename)
    if (force_download == False) & os.path.isfile(path):
        pass
    else:
        path = os.path.join(CELLORACLE_DATA_DIR, filename)
        backup_url = os.path.join(WEB_PAR_DIR, filename)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)

    return pd.read_parquet(path)



def load_Scerevisiae_promoter_base_GRN(version="sacCer3_CisBPv2_fpr2", force_download=False):
    """
    Load Base GRN made from promoter DNA sequence and motif scan.

    Args:

    Returns:
        pandas.dataframe: Base GRN as a matrix.
    """

    options = {"sacCer2_CisBPv2_fpr1": "promoter_base_GRN/sacCer2_TFinfo_dataframe_CisBPv2_fpr1_threshold_10_20210630.parquet",
               "sacCer2_CisBPv2_fpr2": "promoter_base_GRN/sacCer2_TFinfo_dataframe_CisBPv2_fpr2_threshold_10_20210630.parquet",
               "sacCer3_CisBPv2_fpr1": "promoter_base_GRN/sacCer3_TFinfo_dataframe_CisBPv2_fpr1_threshold_10_20210630.parquet",
               "sacCer3_CisBPv2_fpr2": "promoter_base_GRN/sacCer3_TFinfo_dataframe_CisBPv2_fpr2_threshold_10_20210630.parquet",
               }
    if version in options.keys():
        print(f"Loading prebuilt promoter base-GRN. Version: {version}")
        filename = options[version]
    else:
        print(f"Version error. {version} is not in the list.")
        print("Available option: ", list(options.keys()))

    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], filename)
    if (force_download == False) & os.path.isfile(path):
        pass
    else:
        path = os.path.join(CELLORACLE_DATA_DIR, filename)
        backup_url = os.path.join(WEB_PAR_DIR, filename)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)

    return pd.read_parquet(path)

def load_Celegans_promoter_base_GRN(version="ce10_CisBPv2_fpr2", force_download=False):
    """
    Load Base GRN made from promoter DNA sequence and motif scan.

    Args:

    Returns:
        pandas.dataframe: Base GRN as a matrix.
    """

    options = {"ce6_CisBPv2_fpr1": "promoter_base_GRN/ce6_TFinfo_dataframe_CisBPv2_fpr1_threshold_10_20210630.parquet",
               "ce6_CisBPv2_fpr2": "promoter_base_GRN/ce6_TFinfo_dataframe_CisBPv2_fpr2_threshold_10_20210630.parquet",
               "ce10_CisBPv2_fpr1": "promoter_base_GRN/ce10_TFinfo_dataframe_CisBPv2_fpr1_threshold_10_20210630.parquet",
               "ce10_CisBPv2_fpr2": "promoter_base_GRN/ce10_TFinfo_dataframe_CisBPv2_fpr2_threshold_10_20210630.parquet",
               }
    if version in options.keys():
        print(f"Loading prebuilt promoter base-GRN. Version: {version}")
        filename = options[version]
    else:
        print(f"Version error. {version} is not in the list.")
        print("Available option: ", list(options.keys()))

    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], filename)
    if (force_download == False) & os.path.isfile(path):
        pass
    else:
        path = os.path.join(CELLORACLE_DATA_DIR, filename)
        backup_url = os.path.join(WEB_PAR_DIR, filename)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)

    return pd.read_parquet(path)

def load_Pig_promoter_base_GRN(version="Sscrofa11.1_CisBPv2_fpr2", force_download=False):
    """
    Load Base GRN made from promoter DNA sequence and motif scan.

    Args:

    Returns:
        pandas.dataframe: Base GRN as a matrix.
    """

    options = {"Sscrofa11.1_CisBPv2_fpr1": "promoter_base_GRN/Sscrofa11.1_TFinfo_dataframe_CisBPv2_fpr1_threshold_10_20230218.parquet",
               "Sscrofa11.1_CisBPv2_fpr2": "promoter_base_GRN/Sscrofa11.1_TFinfo_dataframe_CisBPv2_fpr2_threshold_10_20230218.parquet",
               }
    if version in options.keys():
        print(f"Loading prebuilt promoter base-GRN. Version: {version}")
        filename = options[version]
    else:
        print(f"Version error. {version} is not in the list.")
        print("Available option: ", list(options.keys()))

    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], filename)
    if (force_download == False) & os.path.isfile(path):
        pass
    else:
        path = os.path.join(CELLORACLE_DATA_DIR, filename)
        backup_url = os.path.join(WEB_PAR_DIR, filename)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)

    return pd.read_parquet(path)



# -*- coding: utf-8 -*-
"""
The :mod:`.network_analysis` module implements Network analysis.

"""

from . import gene_analysis
from . import network_structure_analysis
from .network_construction import get_links
from .use_r_scripts import test_R_libraries_installation, get_R_path, set_R_path, config, _check_R_libraries_installation
from .links_object import load_links, Links
from .network_analysis_utility import transfer_scores_from_links_to_adata, linkList_to_networkgraph, draw_network

__all__ = ["get_links", "gene_analysis", "network_structure_analysis",
           "get_R_path", "set_R_path",
           "load_links", "Links",
           "transfer_scores_from_links_to_adata",
           "linkList_to_networkgraph", "draw_network"]



# -*- coding: utf-8 -*-
"""
The :mod:`.applications` module has several functions that support celloracle.

"""

#from .interactive_simulation_and_plot import Oracle_extended, DEFAULT_PARAMETERS
#from .development_analysis import Oracle_development_module, subset_oracle_for_development_analysiis
from .pseudotime import Pseudotime_calculator
from .differentiation_flow import load_gradient, Gradient_calculator, aggregate_Gradient_objects
from .development_module import Oracle_development_module
from .systematic_analysis_helper import Oracle_systematic_analysis_helper


__all__ = ["Pseudotime_calculator",
            "load_gradient", "Gradient_calculator", "aggregate_Gradient_objects",
           "Oracle_development_module", "Oracle_systematic_analysis_helper"]



# -*- coding: utf-8 -*-
'''
This is a series of custom functions for the inferring of GRN from single cell RNA-seq data.

Codes were written by Kenji Kamimoto.


'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing

import os

from copy import deepcopy
from datetime import datetime
from time import ctime

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from joblib import Parallel, delayed
from tqdm.auto import tqdm

from scipy.stats import ttest_1samp, norm

from ..utility.hdf5_processing import dump_hdf5, load_hdf5

# 0.2. custom libraries
from ..utility import  save_as_pickled_object, standard, intersect
from .regression_models import get_bagging_ridge_coefs as _get_bagging_ridge_coefs
from .regression_models import get_bayesian_ridge_coefs as _get_bayesian_ridge_coefs


RIDGE_SOLVER = "auto"

##############################
### 1.  define main class  ###
##############################

class Net():
    '''
    Net is a custom class for inferring sample-specific GRN from scRNA-seq data.
    This class is used inside the Oracle class for GRN inference.
    This class requires two types of information below.

    (1) Single-cell RNA-seq data:
        The Net class needs processed scRNA-seq data.
        Gene and cell filtering, quality check, normalization, log-transformation (but not scaling and centering) have to be done before starting the GRN calculation with this class.
        You can also use any arbitrary metadata (i.e., mRNA count, cell-cycle phase) for GRN input.

    (2) Potential regulatory connection (or base GRN):
        This method uses the list of potential regulatory TFs as input.
        This information can be calculated from ATAC-seq data using the motif-analysis module.
        If sample-specific ATAC-seq data is not available,
        you can use general TF-binding info derived from public ATAC-seq dataset of various tissue/cell type.

    Attributes:
        linkList (pandas.DataFrame): The results of the GRN inference.
        all_genes (numpy.array): An array of all genes that exist in the input gene expression matrix
        embedding_name (str): The key name name in adata.obsm containing dimensional reduction coordinates
        annotation(dictionary): Annotation. you can add custom annotation.
        coefs_dict (dictionary): Coefs of linear regression.
        stats_dict (dictionary): Statistic values about coefs.
        fitted_genes (list of str): List of genes where the regression model was successfully calculated.
        failed_genes (list of str): List of genes that were not assigned coefs
        cellstate (pandas.DataFrame): A metadata for GRN input
        TFinfo (pandas.DataFrame): Information about potential regulatory TFs.
        gem (pandas.DataFrame): Merged matrix made with gene_expression_matrix and cellstate matrix.
        gem_standerdized (pandas.DataFrame): Almost the same as gem, but the gene_expression_matrix was standardized.
        library_last_update_date (str): Last update date of this code. This info is for code development. It can be deprecated in the future
        object_initiation_date (str): The date when this object was made.

    '''


    #######################################
    ### 1.1. Initialize transNet object ###
    #######################################
    def __init__(self, gene_expression_matrix, gem_standerdized=None, TFinfo_matrix=None, cellstate=None, TFinfo_dic=None, annotation=None, verbose=True):
        '''
        Instantiate Net object

        Args:
            gene_expression_matrix (pandas.DataFrame): scRNA gene expression matrix.
                Column names should be genes. Row names should be cell names.
                Data preprocessing with Seurat or scanpy is recommended.
                See tutorials for more information.

            TFinfo (pandas.DataFrame): information about potential regulatory TFs.

            cellstate (pandas.DataFrame): optional input data.
                Metadata that was acquired during data preprocessing phase can be used for GRN inference.

        '''
        ## 1. Initiate attributes
        if verbose:
            print("initiating Net object ...")
        self.gem = None
        self.gem_standerdized = None
        self.cellstate = None
        self.all_genes = None
        self.TFinfo = None
        self.TFdict = {}
        self.linkList = None
        self.annotation = annotation
        self.fitted_genes = []
        self.failed_genes = []
        self.coefs_dict = {}
        self.stats_dict = {}
        self.library_last_update_date = ctime(os.path.getmtime(__file__))
        self.object_initiation_date = datetime.now().ctime()
        if self.annotation is None:
            self.annotation = {}

        ## 2. Process attributes
        # 2.1. gene expression matrix, standerdized_gem, cellstate
        self.gem = gene_expression_matrix.copy()
        self.gem.index.name = None
        self.gem.columns.name = None

        if gem_standerdized is None:
            self.gem_standerdized = standard(self.gem)
        else:
            self.gem_standerdized = gem_standerdized.copy()
            self.gem_standerdized.index.name = None
            self.gem_standerdized.columns.name = None

        if not cellstate is None:
            self.cellstate = cellstate.copy()
            self.cellstate.index.name = None
            self.cellstate.columns.name = None
            print(f"cellstate shape: {self.cellstate.shape}")

            if not self.gem.shape[0] == self.cellstate.shape[0]:
                raise ValueError("cellnumber in GEM and cellstate shold be same.")
            else:
                self.gem = pd.concat([self.gem, self.cellstate], axis=1)
                self.gem_standerdized = pd.concat([self.gem_standerdized,
                                                   self.cellstate], axis=1)
                self.annotation["cellstate"] = list(self.cellstate.columns)
        self.gem_standerdized = self.gem_standerdized.astype("float32")
        self.gem = self.gem.astype("float32")


        if verbose:
            print(f"gem_shape: {self.gem.shape}")

        # 2.2. Gene list
        self.all_genes = np.unique(self.gem.columns)

        # 2.3. TF_information
        if not TFinfo_matrix is None:
            self.TFinfo = TFinfo_matrix.copy()
            self.TFinfo.index.name = None
            self.TFinfo.columns.name = None
            self.TFinfo = self.TFinfo.reset_index(drop=True)


            tmp = self.TFinfo.copy()
            tmp = tmp.drop(["peak_id"], axis=1)
            tmp = tmp.groupby(by="gene_short_name").sum()
            self.TFdict = dict(tmp.apply(lambda x: x[x > 0].index.values, axis=1))
            if verbose:
                print(f"TF info shape: {self.TFinfo.shape}")


        if not TFinfo_dic is None:
            self.TFdict.update(TFinfo_dic)

        if verbose:
            print("initiation completed.")

    def copy(self):
        """
        Deepcopy itself
        """
        return deepcopy(self)

    def addTFinfo_matrix(self, TFinfo_matrix):
        '''
        Load TF info dataframe.

        Args:
            TFinfo (pandas.DataFrame): information about potential regulatory TFs.

        '''
        self.TFinfo = TFinfo_matrix.copy()
        tmp = TFinfo_matrix.copy()
        tmp = tmp.drop(["peak_id"], axis=1)
        tmp = tmp.groupby(by="gene_short_name").sum()
        self.TFdict = dict(tmp.apply(lambda x: x[x > 0].index.values, axis=1))

    def updateTFinfo_dictionary(self, TFdict):
        '''
        Update TF info matrix

        Args:
            TFdict (dictionary): A python dictionary in which a key is Target gene, value are potential regulatory genes for the target gene.

        '''
        self.TFdict.update(TFdict)

    def addTFinfo_dictionary(self, TFdict):
        """
        Add a new TF info to pre-exiting TFdict.

        Args:
            TFdict (dictionary): python dictionary of TF info.
        """

        for tf in TFdict:
            if tf in self.TFdict.keys():
                targets = self.TFdict[tf]
                targets = list(TFdict[tf]) + list(targets)
                targets = np.unique(targets)
                self.TFdict.update({tf: targets})
            else:
                self.TFdict.update({tf: TFdict[tf]})



    def addAnnotation(self, annotation_dictionary):
        '''
        Add a new annotation.

        Args:
            annotation_dictionary (dictionary): e.g. {"sample_name": "NIH 3T3 cell"}
        '''

        self.annotation.update(annotation_dictionary)



    ##############################################################
    ## 2.1. Make and fit Regularized Linear model to get Coefs  ##
    ##############################################################


    def fit_All_genes_parallel(self, bagging_number=200, scaling=True, log=None, verbose=10):
        """
        IMPORTANT: this function being debugged and is currently unavailable.

        Make ML models for all genes.
        The calculation will be performed in parallel using joblib parallel module.

        Args:
            bagging_number (int): The number of estimators for bagging.
            scaling (bool): Whether or not to scale regulatory gene expression values.
            log (logging object): log object to output log
            verbose (int): verbose for joblib parallel
        """

        # prepare for parallel calculation
        def process(target_gene):

            coefs = _get_bagging_ridge_coefs(target_gene=target_gene,
                                             gem=self.gem,
                                             gem_scaled=self.gem_standerdized,
                                             TFdict=self.TFdict,
                                             cellstate=self.cellstate,
                                             bagging_number=bagging_number,
                                             scaling=scaling,
                                             n_jobs=1,
                                             alpha=alpha)
            if type(coefs) is int:
                fitted_gene = "na"
                failed_gene = target_gene
            else:
                fitted_gene = target_gene
                failed_gene = "na"
            return _get_stats_df_bagging_ridge(coefs), fitted_gene, failed_gene

        # do parallel calculation with joblib
        genes = np.array(intersect(self.all_genes, self.TFdict.keys()))
        results = Parallel(n_jobs=-1,
                           backend="threading",
                           verbose=verbose)([delayed(process)(i) for i in genes])

        # order and save results
        results = np.array(results)
        values, fitted_genes, failed_genes = results[:, 0], results[:, 1], results[:, 2]
        self.stats_dict = dict(zip(fitteg_genes, values))
        self.stats_dict.pop("na")
        fitted_genes = np.delete(fitted_genes, np.where(fitted_genes == "na"))
        failed_genes = np.delete(failed_genes, np.where(failed_genes == "na"))
        self.fitted_genes = fitted_genes
        self.failed_genes = failed_genes

    def fit_All_genes(self, bagging_number=200, scaling=True, model_method="bagging_ridge",
                      command_line_mode=False, log=None, alpha=1, verbose=True, n_jobs=-1):
        """
        Make ML models for all genes.
        The calculation will be performed in parallel using scikit-learn bagging function.
        You can select a modeling method (bagging_ridge or bayesian_ridge).  This calculation usually takes a long time.

        Args:
            bagging_number (int): The number of estimators for bagging.
            scaling (bool): Whether or not to scale regulatory gene expression values.
            model_method (str): ML model name. Please select either "bagging_ridge" or "bayesian_ridge"
            command_line_mode (bool): Please select False if the calculation is performed on jupyter notebook.
            log (logging object): log object to output log
            alpha (int) : Strength of regularization.
            verbose (bool): Whether or not to show a progress bar.
            n_jobs (int): Number of cpu cores for parallel calculation. -1 means using all available cores.
        """
        self.fit_genes(target_genes=self.all_genes,
                       bagging_number=bagging_number,
                       scaling=scaling,
                       model_method=model_method,
                       save_coefs=False,
                       command_line_mode=command_line_mode,
                       log=log,
                       alpha=alpha,
                       verbose=verbose,
                       n_jobs=n_jobs)

    def fit_genes(self, target_genes, bagging_number=200, scaling=True, model_method="bagging_ridge",
                  save_coefs=False, command_line_mode=False, log=None, alpha=1, verbose=True, n_jobs=-1):
        """
        Make ML models for genes of interest.
        This calculation will be performed in parallel using scikit-learn's bagging function.
        You can select a modeling method; Please chose either bagging_ridge or bayesian_ridge.

        Args:
            target_genes (list of str): gene list
            bagging_number (int): The number of estimators for bagging.
            scaling (bool): Whether or not to scale regulatory gene expression values.
            model_method (str): ML model name. Please select either "bagging_ridge" or "bayesian_ridge"
            save_coefs (bool): Whether or not to store details of coef values in bagging model.
            command_line_mode (bool): Please select False if the calculation is performed on jupyter notebook.
            log (logging object): log object to output log
            alpha (int) : Strength of regularization.
            verbose (bool): Whether or not to show a progress bar.
            n_jobs (int): Number of cpu cores for parallel calculation.  -1 means using all available cores.

        """
        genes = np.array(intersect(target_genes, self.TFdict.keys()))
        genes = np.array(intersect(genes, self.all_genes))
        if verbose:
            #print(f"method: {model_method}")
            if model_method == "bagging_ridge":
                #print(f"alpha: {alpha}")
                pass

        if command_line_mode:

            N = len(genes)
            log_step = 10

            if model_method == "bagging_ridge":

                for i, target_gene in enumerate(genes):
                    coefs = _get_bagging_ridge_coefs(target_gene=target_gene,
                                                     gem=self.gem,
                                                     gem_scaled=self.gem_standerdized,
                                                     TFdict=self.TFdict,
                                                     cellstate=self.cellstate,
                                                     bagging_number=bagging_number,
                                                     scaling=scaling,
                                                     n_jobs=n_jobs,
                                                     alpha=alpha,
                                                     solver=RIDGE_SOLVER)

                    if isinstance(coefs, int):
                        self.failed_genes.append(target_gene)

                    else:
                        self.fitted_genes.append(target_gene)
                        self.stats_dict[target_gene] = _get_stats_df_bagging_ridge(coefs)
                        if save_coefs:
                            self.coefs_dict[target_gene] = coefs

                    if i/N*100 >= log_step:
                        #print(f"{datetime.now().ctime()}: {log_step}% processed..")
                        log.info(f"{log_step}% processed..")
                        log_step += 10

            elif model_method == "bayesian_ridge":

                for i, target_gene in enumerate(genes):
                    coef_mean, coef_variance, coef_names = \
                        _get_bayesian_ridge_coefs(target_gene=target_gene,
                                                  gem=self.gem,
                                                  gem_scaled=self.gem_standerdized,
                                                  TFdict=self.TFdict,
                                                  cellstate=self.cellstate,
                                                  scaling=True)

                    if isinstance(coef_mean, int):
                        self.failed_genes.append(target_gene)

                    else:
                        self.fitted_genes.append(target_gene)
                        stats_df = _get_stats_df_from_bayesian_ridge(coef_mean=coef_mean,
                                                                     coef_variance=coef_variance,
                                                                     coef_names=coef_names)
                        self.stats_dict[target_gene] = stats_df

                    if i/N*100 >= log_step:
                        #print(f"{datetime.now().ctime()}: {log_step}% processed..")
                        log.info(f"{log_step}% processed..")
                        log_step += 10

            log.info("100% processed..")
            #print(f"{datetime.now().ctime()}: 100% processed..")

        else:

            if model_method == "bagging_ridge":
                if verbose:
                    loop = tqdm(genes)
                else:
                    loop = genes

                for target_gene in loop:
                    coefs = _get_bagging_ridge_coefs(target_gene=target_gene,
                                                     gem=self.gem,
                                                     gem_scaled=self.gem_standerdized,
                                                     TFdict=self.TFdict,
                                                     cellstate=self.cellstate,
                                                     bagging_number=bagging_number,
                                                     scaling=scaling,
                                                     n_jobs=n_jobs,
                                                     alpha=alpha,
                                                     solver=RIDGE_SOLVER)

                    if isinstance(coefs, int):
                        self.failed_genes.append(target_gene)

                    else:
                        self.fitted_genes.append(target_gene)
                        self.stats_dict[target_gene] = _get_stats_df_bagging_ridge(coefs)

            elif model_method == "bayesian_ridge":

                for target_gene in tqdm(genes):
                    coef_mean, coef_variance, coef_names = \
                        _get_bayesian_ridge_coefs(target_gene=target_gene,
                                                  gem=self.gem,
                                                  gem_scaled=self.gem_standerdized,
                                                  TFdict=self.TFdict,
                                                  cellstate=self.cellstate,
                                                  scaling=True)

                    if isinstance(coef_mean, int):
                        self.failed_genes.append(target_gene)

                    else:
                        self.fitted_genes.append(target_gene)
                        stats_df = _get_stats_df_from_bayesian_ridge(coef_mean=coef_mean,
                                                                     coef_variance=coef_variance,
                                                                     coef_names=coef_names)
                        self.stats_dict[target_gene] = stats_df

    #########################
    ### 3.1. Plot results ###
    #########################

    def plotCoefs(self, target_gene, sort=True, threshold_p=None):
        """
        Plot the distribution of Coef values (network edge weights).

        Args:
            target_gene (str): gene name
            sort (bool): Whether or not to sort genes by its strength
            bagging_number (int): The number of estimators for bagging.
            threshold_p (float): the threshold for p-values. TFs will be filtered based on the p-value.
                if None, no filtering is applied.
        """
        coefs = self.coefs_dict[target_gene].copy()
        stat = self.stats_dict[target_gene].copy()

        if not threshold_p is None:
            sig_TFs = (stat.percentile_of_zero > (100-threshold_p*100)) |\
                (stat.percentile_of_zero < (threshold_p*100))
            sig_TFs = stat.index.values[sig_TFs]
            coefs = coefs[sig_TFs]
            stat = stat.loc[sig_TFs]

        if sort:
            sorted_columns = stat.sort_values(by="positive_score",
                                              ascending=False).index.values
            coefs = coefs[sorted_columns]

        melted = _get_melted_df(coefs)
        sns.boxplot(data=melted, x="variable", y="value")
        plt.title(target_gene)
        plt.axhline(0)
        plt.xticks(range(len(coefs.columns)), coefs.columns, rotation="vertical")
    #############################
    ### 4.1. Process results  ###
    #############################

    # get linkList
    def updateLinkList(self, verbose=True):
        """
        Update LinkList.
        LinkList is a data frame that store information about inferred GRNs.

        Args:
            verbose (bool): Whether or not to show a progress bar

        """
        if not self.fitted_genes: # if the sequence is empty
            print("No model found. Do fit first.")

        linkList = []
        if verbose:
            loop = tqdm(np.unique(self.fitted_genes))
        else:
            loop = np.unique(self.fitted_genes)

        for i in loop:
            linkList.append(_stats2LinkList(i, stat_df=self.stats_dict[i]))

        linkList = pd.concat(linkList, axis=0)
        linkList = linkList.reset_index(drop=True)

        self.linkList = linkList





    ####################################
    ### 5.1 Save and Output results  ###
    ####################################


    def _save_as_pickle(self, folder):

        """
        Save itself as a pickled data.
        The pickled data may be large.
        Please use "load_pickled_object" function in utility module to load saved file.

        Args:
            foder (str): path to save the Net object.
                The folder will be created if the folder does not exist.

        """

        os.makedirs(folder, exist_ok=True)
        save_as_pickled_object(self, os.path.join(folder, "transNet.pickle"))

        print(f"file saved at: {folder}")


    def _save_as_parquet(self, folder=None):
        """
        Save itself. DataFrame in this object will be saved as a parquet file.
        Please use "load_compressed_TN_object" function in the network module to load the saved file.

        Args:
            folder (str): path to save the Net object.
                The folder will be created if the folder does not exist.

        """

        os.makedirs(folder, exist_ok=True)

        tmp_self = self.copy()

        tmp_self.gem.to_parquet(folder + "/gem.parquet")
        tmp_self.gem = None

        tmp_self.gem_standerdized.to_parquet(folder + "/gem_standerdized.parquet")
        tmp_self.gem_standerdized = None

        if not self.linkList is None:
            tmp_self.linkList.to_parquet(folder + "/linkList.parquet")
            tmp_self.linkList = None

        if not self.TFinfo is None:
            tmp_self.TFinfo.to_parquet(folder + "/TFinfo.parquet")
            tmp_self.TFinfo = None

        if not self.cellstate is None:
            tmp_self.cellstate.to_parquet(folder + "/cellstate.parquet")
            tmp_self.cellstate = None


        save_as_pickled_object(tmp_self, folder + "/transNet.pickle")

        print(f"file saved at: {folder}")


    def to_hdf5(self, file_path):
        """
        Save object as hdf5.

        Args:
            file_path (str): file path to save file. Filename needs to end with '.celloracle.net'
        """
        if file_path.endswith(".celloracle.net"):
            pass
        else:
            raise ValueError("Filename needs to end with '.celloracle.net'")

        compression_opts = 7
        dump_hdf5(obj=self, filename=file_path,
                  data_compression=compression_opts,  chunks=(2048, 2048),
                  noarray_compression=compression_opts, pickle_protocol=4)

####################################################
### 2. Define functions for transNet calculation ###
####################################################



# this is a function to convert DataFrame to visualize itself by seaborn plot function.
def _get_melted_df(df):
    d_ = df.copy()
    d_["index_name"] = df.index
    melted = pd.melt(d_, id_vars=['index_name'], var_name='variable', value_name='value')
    return melted

# this function process coefs to get several statistical values
import warnings

def _get_stats_df_bagging_ridge(df):

    if isinstance(df, int):
        return 0

    mean = df.mean()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        p = df.apply(lambda x: ttest_1samp(x.dropna(), 0)[1])
    neg_log_p = -np.log10(p.fillna(1))

    result = pd.concat([mean, mean.abs(),
                        p, neg_log_p, #positive_score, negative_score
                        ], axis=1, sort=False)
    result.columns = ["coef_mean", "coef_abs", "p", "-logp",
                      #"positive_score", "negative_score"
                      ]

    return result

def _get_stats_df_from_bayesian_ridge(coef_mean, coef_variance, coef_names):

    coef_abs = np.abs(coef_mean)
    p = norm.cdf(x=0, loc=coef_abs, scale=np.sqrt(coef_variance))*2
    neg_log_p = -np.log(p)

    stats_df = pd.DataFrame({"coef_mean": coef_mean,
                             "coef_abs": coef_abs,
                             "coef_variance": coef_variance,
                             "p": p,
                             "-logp": neg_log_p},
                             index = coef_names)
    return stats_df

def _stats2LinkList(tg, stat_df):

    links = pd.DataFrame({"source": stat_df.index.values,
                          "target": np.repeat(tg, len(stat_df))})
    linkList = pd.concat([links, stat_df.reset_index(drop=True)], axis=1)

    return linkList



# -*- coding: utf-8 -*-
'''
This is a series of custom functions for the inferring of GRN from single cell RNA-seq data.

Codes were written by Kenji Kamimoto.


'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing
import glob
import os

import numpy as np
import pandas as pd

# 0.3. custom libraries
from ..utility import load_pickled_object
from .net_core import Net

from velocyto.serialization import dump_hdf5, load_hdf5

##############################
### 1.  define main class  ###
##############################


def load_net(file_path):
    return load_hdf5(filename=file_path, obj_class=Net)


def load_net_from_patquets(folder_path):
    """
    Load a Net object that was saved with "save_as_compressed" function.

    Args:
        folder_path (str): Folder path
    """


    obj = load_pickled_object(os.path.join(folder_path, "transNet.pickle"))

    compressed_files_path = glob.glob(os.path.join(folder_path, "*.parquet"))
    compressed_file_name = [i.split("/")[-1] for i in  compressed_files_path]

    for name, path in zip(compressed_file_name, compressed_files_path):
        if name == "gem.parquet":
            obj.gem = pd.read_parquet(path)

        if name == "gem_standerdized.parquet":
            obj.gem_standerdized = pd.read_parquet(path)

        if name == "TFinfo.parquet":
            obj.TFinfo = pd.read_parquet(path)

        if name == "cellstate.parquet":
            obj.cellstate = pd.read_parquet(path)

        if name == "linkList.parquet":
            obj.linkList = pd.read_parquet(path)
    return obj



def getDF_TGxTF(net_object, value_of_interest):
    """
    Extract inferred GRN information and return as a pandas.DataFrame.
    The results was converted as Target gene x TF.


    Args:
        net_object (Net): Net object. GRN calculation have to be done in this object.
        value_of_interest (str): Kind of information to extract.

    """
    if not net_object.fitted_genes:
        print("No model found. Do fit first.")
        return None

    li = []
    genes = np.unique(net_object.fitted_genes)
    for i in genes:
        li.append(net_object.stats_dict[i][[value_of_interest]])

    df = pd.concat(li, axis=1, sort=False)
    df.columns = genes
    df = df.transpose().fillna(0)

    return df

def getDF_peakxTF(net_object, stat_value_of_interest):
    """
    Extract inferred GRN information and return as a pandas.DataFrame.
    The results was converted as peak (genome locus) x TF.

    Args:
        net_object (Net): Net object. GRN calculation have to be done in this object.
        value_of_interest (str): Kind of information to extract.

    """
    # load DF
    DF_grn = getDF_TGxTF(net_object, stat_value_of_interest)
    DF_tfi = trans_net_object.TFinfo.copy()

    # get index and column info
    index_info = DF_tfi[["peak_id", "gene_short_name"]]
    column_info = DF_grn.columns.values

    # convert DF_grn
    DF_grn = DF_grn.reindex(DF_tfi.gene_short_name.values).fillna(0)
    DF_grn = DF_grn.reset_index(drop=True)

    # convert DF_tfi
    DF_tfi = DF_tfi[column_info]

    merge = DF_grn * DF_tfi
    merge = pd.concat([index_info, merge], axis=1)

    merge = merge.set_index("peak_id")
    return merge



import os
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc

plt.rcParams["savefig.dpi"] = 300
plt.rcParams["figure.figsize"] = [6, 4.5]


def main():
    test_tutorial_scRNAseq_data_processing()

def test_tutorial_scRNAseq_data_processing():
    # Start analysis
    import time
    start = time.time()



    # 1. Load data
    print("\n1. Load data")
    # Download dataset. You can change the code blow to use your data.
    adata = sc.datasets.paul15()

    # 2. Filtering
    print("\n2. Filtering")
    # Only consider genes with more than 1 count
    sc.pp.filter_genes(adata, min_counts=1)


    # 3. Normalizatioin
    print("\n3. Normalization")
    # Normalize gene expression matrix with total UMI count per cell
    sc.pp.normalize_per_cell(adata, key_n_counts='n_counts_all')

    # 4. Identificatin of highly variable genes
    print("\n4. Identificatin of highly variable genes")
    # Select top 2000 highly-variable genes
    filter_result = sc.pp.filter_genes_dispersion(adata.X,
                                                  flavor='cell_ranger',
                                                  n_top_genes=2000,
                                                  log=False)
    # Subset the genes
    adata = adata[:, filter_result.gene_subset]
    # Renormalize after filtering
    sc.pp.normalize_per_cell(adata)

    # 5. Log transformation
    print("\n5. Log transformation")
    # keep raw cont data before log transformation
    adata.raw = adata
    adata.layers["raw_count"] = adata.raw.X.copy()
    # Log transformation and scaling
    sc.pp.log1p(adata)
    sc.pp.scale(adata)

    # 6. PCA and neighbor calculations
    print("\n6. PCA and neighbor calculations")
    # PCA
    sc.tl.pca(adata, svd_solver='arpack')
    # Diffusion map
    sc.pp.neighbors(adata, n_neighbors=4, n_pcs=20)
    sc.tl.diffmap(adata)
    # Calculate neihbors again based on diffusionmap
    sc.pp.neighbors(adata, n_neighbors=10, use_rep='X_diffmap')

    # 7. Cell clustering
    print("\n7. Cell clustering")
    sc.tl.louvain(adata, resolution=0.8)

    # 8. Dimensionality reduction using PAGA and force-directed graphs
    print("\n8. Dimensionality reduction using PAGA and force-directed graphs")
    # PAGA graph construction
    sc.tl.paga(adata, groups='louvain')
    plt.rcParams["figure.figsize"] = [6, 4.5]
    sc.pl.paga(adata)
    sc.tl.draw_graph(adata, init_pos='paga', random_state=123)
    sc.pl.draw_graph(adata, color='louvain', legend_loc='on data')

    # 9. Checkdata
    print("\n9. Check data")
    plt.rcParams["figure.figsize"] = [4.5, 4.5]
    markers = {"Erythroids":["Gata1", "Klf1", "Gypa", "Hba-a2"],
               "Megakaryocytes":["Itga2b", "Pbx1", "Sdpr", "Vwf"],
                "Granulocytes":["Elane", "Cebpe", "Ctsg", "Mpo", "Gfi1"],
                "Monocytes":["Irf8", "Csf1r", "Ctsg", "Mpo"],
                "Mast_cells":["Cma1", "Gzmb", "Kit"],
                "Basophils":["Mcpt8", "Prss34"]
                }
    for cell_type, genes in markers.items():
        print(f"marker gene of {cell_type}")
        sc.pl.draw_graph(adata, color=genes, use_raw=False, ncols=2)
        plt.show()

    adata.write_h5ad("Paul_etal_15.h5ad")

    # 6. Remove files
    print("Remove files")
    for i in ["Paul_etal_15.h5ad"]:
        os.remove(i)
    os.remove("data/paul15/paul15.h5")
    os.removedirs("data/paul15")

    elapsed = time.time() - start

    print("Success")
    print(f"Total time: {elapsed/60:.3g} minutes")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-



import io
import logging
import os
import pickle
import subprocess
import sys

import pandas as pd
import numpy as np
from tqdm.auto import tqdm
import matplotlib.pyplot as plt
import h5py


from .utility import Oracle_data_strage
from .development_analysis import Oracle_development_module

from ..trajectory.oracle_core import Oracle

#import scanpy as sc
#import seaborn as sns
#from ipywidgets import interactive
import h5py


DEFAULT_PARAMETERS = {"n_neighbors": 200,
                      "quiver_scale": 30, "quiver_scale_grid": 1.5,
                      "min_mass": 1.0, "n_grid": 40,
                      "min_magnitude": None,
                      "sampled_fraction": 0.5,
                      "n_propagation": 3,
                      "n_steps_for_mc": 20,
                      "gene": "Gata1"}


COMMON_ATTRS = ["embedding", "colorandum", "cluster_column_name"]


class Oracle_extended(Oracle_data_strage, Oracle_development_module):

    def __init__(self, oracle, hdf_path, mode, obsm_key="X_umap"):
        self.oracle = oracle
        self.gene = None
        self.n_neighbors = None
        self.n_grid = None
        self.names = []
        self.obsm_key = obsm_key

        self.set_hdf_path(path=hdf_path, create_if_not_exist=True)

        if mode == "write":
            self.oracle.embedding = self.oracle.adata.obsm[self.oracle.embedding_name]
            self.save_data(oracle=self.oracle, place=f"common", attributes=COMMON_ATTRS)

        elif mode == "read":
            self.load_data(oracle=self.oracle, place=f"common", attributes=COMMON_ATTRS)
            self.oracle.cluster_column_name = str(self.oracle.cluster_column_name)

    # Module for perturb simulation
    def load_simulation(self, gene=DEFAULT_PARAMETERS["gene"]):
        print("Loading data ..")
        self.load_data(oracle=self.oracle, place=f"simulation/{gene}",
                       attributes=["delta_embedding", "delta_embedding_random"])

        self.load_dfs(oracle=self.oracle, place=f"mcmc/{gene}",
                      attributes=["mcmc_transition", "mcmc_transition_random"])

        self.oracle.corrcoef_random = "dummy"
        self.gene = gene
        print(f"Data loaded. Perturbed gene: {gene}")

    def interactive_simulation(self, gene,
                               n_neighbors=DEFAULT_PARAMETERS["n_neighbors"],
                               sampled_fraction=DEFAULT_PARAMETERS["sampled_fraction"],
                               n_propagation=DEFAULT_PARAMETERS["n_propagation"],
                               n_steps_for_mc=DEFAULT_PARAMETERS["n_steps_for_mc"],
                               save=True):
        """


        """
        print("Calculating ...")
        # Enter perturbation conditions to simulate signal propagation after the perturbation.
        self.oracle.simulate_shift(perturb_condition={gene: 0},
                                   ignore_warning=True,
                                   n_propagation=n_propagation)

        # Get transition probability
        self.oracle.estimate_transition_prob(n_neighbors=n_neighbors, knn_random=True, sampled_fraction=sampled_fraction)

        # Calculate embedding
        self.oracle.calculate_embedding_shift(sigma_corr = 0.05)

        # Do Markov simulation
        self.oracle.run_markov_chain_simulation(n_steps=n_steps_for_mc, n_duplication=1)
        self.oracle.get_mcmc_cell_transition_table(end=n_steps_for_mc)


        self.gene = gene
        self.n_neighbors = n_neighbors

        if save:
            self.save_data(oracle=self.oracle, place=f"simulation/{gene}",
                           attributes=["delta_embedding", "delta_embedding_random"])


            self.save_dfs(oracle=self.oracle, place=f"mcmc/{gene}",
                          attributes=["mcmc_transition", "mcmc_transition_random"])

            print("Results were saved to hdf5 file.")

        print("finished")


    def interactive_plot_quiver(self, quiver_scale=DEFAULT_PARAMETERS["quiver_scale"]):

        plt.figure(None,(6,6))
        plt.scatter(self.oracle.embedding[:, 0], self.oracle.embedding[:, 1],
                c="0.8", alpha=0.2, s=38, edgecolor=(0,0,0,1), lw=0.3, rasterized=True)

        quiver_kwargs=dict(headaxislength=7, headlength=11, headwidth=8,
                       linewidths=0.35, width=0.0045,edgecolors="k",
                       color=self.oracle.colorandum, alpha=1)
        plt.quiver(self.oracle.embedding[:, 0], self.oracle.embedding[:, 1],
                   self.oracle.delta_embedding[:, 0], self.oracle.delta_embedding[:, 1],
                   scale=quiver_scale, **quiver_kwargs)

        plt.title(f"Knockout simulation: {self.gene}")
        plt.axis("off")

    def interactive_plot_grid(self,
                              quiver_scale=DEFAULT_PARAMETERS["quiver_scale_grid"],
                              min_mass=DEFAULT_PARAMETERS["min_mass"],
                              #min_magnitude=DEFAULT_PARAMETERS["min_magnitude"],
                              n_grid=DEFAULT_PARAMETERS["n_grid"],
                              plot_random=False):
        # Plot whole graph
        if plot_random:
            plt.figure(None,(13,6))
        else:
            plt.figure(None,(6,6))

        if self.n_grid is None:
            self.oracle.calculate_grid_arrows(smooth=0.8, steps=(n_grid, n_grid), n_neighbors=200)
        else:
            if n_grid != self.n_grid:
                self.oracle.calculate_grid_arrows(smooth=0.8, steps=(n_grid, n_grid), n_neighbors=200)

        plt.title(f"Perturb simulation: {self.gene}")

        self.oracle.plot_grid_arrows(quiver_scale=quiver_scale,
                                scatter_kwargs_dict={"alpha":0.35, "lw":0.35,
                                                      "edgecolor":"0.4", "s":38,
                                                      "rasterized":True},
                                min_mass=min_mass,
                                min_magnitude=None,
                                angles='xy', scale_units='xy',
                                headaxislength=2.75,
                                headlength=5, headwidth=4.8, minlength=1.5,
                                plot_random=plot_random, scale_type="relative")




    def _plot_quiver_for_a_cluster(self, cluster_column_name, cluster, quiver_scale, plot_whole_cells=False):

        ix_choice = _get_ix_for_a_cluster(self.oracle, cluster_column_name, cluster)

        if plot_whole_cells:

            plt.scatter(self.oracle.embedding[:, 0], self.oracle.embedding[:, 1],
                        c="lightgray", alpha=1, s=38, lw=0.3, rasterized=True)

        plt.scatter(self.oracle.embedding[ix_choice, 0], self.oracle.embedding[ix_choice, 1],
                    c="0.8", alpha=0.2, s=38, edgecolor=(0,0,0,1), lw=0.3, rasterized=True)



        quiver_kwargs=dict(headaxislength=7, headlength=11, headwidth=8,
                           linewidths=0.35, width=0.0045,edgecolors="k",
                           color=self.oracle.colorandum[ix_choice], alpha=1)
        plt.quiver(self.oracle.embedding[ix_choice, 0], self.oracle.embedding[ix_choice, 1],
                   self.oracle.delta_embedding[ix_choice, 0],
                   self.oracle.delta_embedding[ix_choice, 1],
                   scale=quiver_scale, **quiver_kwargs)

        plt.axis("off")

    def interactive_plot_quiver_for_a_cluster(self, cluster_column_name, cluster,
                                              quiver_scale=DEFAULT_PARAMETERS["quiver_scale"]):
        plt.figure(figsize=(13, 6))

        plt.title(f"Knockout simulation: {self.gene}")
        plt.subplot(1, 2, 1)
        self._plot_quiver_for_a_cluster(cluster_column_name=cluster_column_name,
                                        cluster=cluster, quiver_scale=quiver_scale,
                                        plot_whole_cells=False)

        plt.subplot(1, 2, 2)
        self._plot_quiver_for_a_cluster(cluster_column_name=cluster_column_name,
                                        cluster=cluster, quiver_scale=quiver_scale,
                                        plot_whole_cells=True)

    # Development analysis module
    def save_development_analysis_results(self, gene, cluster_column_name, cluster):

        self.save_data(oracle=self.oracle_dev,
                       place=f"dev_analysis/{cluster_column_name}/{cluster}/{gene}",
                       attributes=["embedding", "pseudotime", "mass_filter",
                                   "flow_grid", "flow", "flow_norm_rndm",
                                   "new_pseudotime", "gradient", "inner_product", "stage", "stage_grid"])

        self.save_dfs(oracle=self.oracle_dev,
                      place=f"inner_product/{cluster_column_name}/{cluster}/{gene}",
                      attributes=["inner_product_stats", "inner_product_stats_grouped"])


        print("Results were saved to hdf5 file.")

    def load_development_analysis_results(self, gene, cluster_column_name, cluster):

        if not hasattr(self, "oracle_dev"):
            setattr(self, "oracle_dev", Oracle())

        print("Loading data ..")
        self.load_data(oracle=self.oracle_dev,
                       place=f"dev_analysis/{cluster_column_name}/{cluster}/{gene}",
                       attributes=["embedding", "pseudotime", "mass_filter",
                                   "flow_grid", "flow", "flow_norm_rndm",
                                   "new_pseudotime", "gradient",
                                   "inner_product", "stage", "stage_grid"])

        self.load_dfs(oracle=self.oracle_dev,
                      place=f"inner_product/{cluster_column_name}/{cluster}/{gene}",
                      attributes=["inner_product_stats", "inner_product_stats_grouped"])

        self.gene = gene
        self.oracle_dev.cluster_loaded = cluster
        self.oracle_dev.cluster_column_name_loaded = cluster_column_name

        print(f"Data loaded. Gene: {gene}")


def _get_ix_for_a_cluster(oracle, cluster_column_name, cluster):
    ix = np.arange(oracle.adata.shape[0])[oracle.adata.obs[cluster_column_name] == cluster]
    return ix



# -*- coding: utf-8 -*-
'''
This is a series of custom functions for the inferring of GRN from single cell RNA-seq data.

Codes were written by Kenji Kamimoto.


'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
#import seaborn as sns

import sys
import os
import glob
from copy import deepcopy
import pickle
from tqdm.auto import tqdm
import seaborn as sns


from joblib import dump, load


##########################
### internal functions ###
##########################

def _plot_base_cartography(z_min, z_max, args_line={}):


    default = {"linestyle": "dashed",
               "alpha": 0.5,
               "c": "gray"}
    default.update(args_line)
    args_line = default.copy()

    plt.plot([-0.05, 1.05], [2.5, 2.5], **args_line)
    plt.plot([0.05, 0.05], [z_min, 2.5], **args_line)
    plt.plot([0.62, 0.62], [z_min, 2.5], **args_line)
    plt.plot([0.8, 0.8], [z_min, 2.5], **args_line)
    plt.plot([0.3, 0.3], [2.5, z_max + 0.5], **args_line)
    plt.plot([0.75, 0.75], [2.5, z_max + 0.5], **args_line)

    plt.xlabel("Participation coefficient (P)")
    plt.ylabel("Whithin-module degree (z)")

    plt.xlim([-0.1, 1.1])


def __plot_goi(data, goi, args_annot, scatter=False):


    x = data.loc[goi, "p"]
    y = data.loc[goi, "z"]
    if scatter:
        plt.scatter(x, y, c="none", edgecolor="black")

    __plot_annotations(x, y, goi, args=args_annot)


@np.vectorize
def __plot_annotations(x, y, label, x_shift=0.05, y_shift=0.05,
                       args={}):

    args_annot = {"size": 10, "color": "black"}
    args_annot.update(args)

    arrow_dict = {"width": 0.5, "headwidth": 0.5, "headlength": 1, "color": "black"}

    plt.annotate(label, xy=(x, y), xytext=(x+x_shift, y+y_shift),
                 arrowprops=arrow_dict, **args_annot)



def add_label_cartography(data):

    # categorize nodes
    R1_nodes = data[(data.p <= 0.05) & (data.z < 2.5)].index
    R2_nodes = data[(data.p > 0.05) & (data.p <= 0.62) & (data.z < 2.5)].index
    R3_nodes = data[(data.p > 0.62) & (data.p <= 0.80) & (data.z < 2.5)].index
    R4_nodes = data[(data.p > 0.80) & (data.z < 2.5)].index
    R5_nodes = data[(data.p <= 0.30) & (data.z >= 2.5)].index
    R6_nodes = data[(data.p > 0.30) & (data.p <= 0.75) & (data.z >= 2.5)].index
    R7_nodes = data[(data.p > 0.75) & (data.z >= 2.5)].index

    node_kind = ["R1: Ultra peripheral", "R2: Peripheral", "R3: Non-hub",
                 "R4: Non-hub kinless", "R5: Provincial",
                 "R6: Connector hubs", "R7: Kinless hubs"]

    data["label_cartography"] = "0"

    for i in range(1, 8):
        data.loc[eval(f"R{i}_nodes"), "label_cartography"] = node_kind[(i-1)]

    return data


##################################
### functions for making plots ###
##################################

### scatter plots
def plot_cartography_color_scatter(data_, gois=None, args_dot={}, args_line={},
                     args_annot={}):

    data = data_.copy()
    data.columns = ["z", "p"]
    data = add_label_cartography(data)

    _plot_base_cartography(z_min=0, z_max=data.z.max(),
                           args_line=args_line)

    labels = list(data.label_cartography.unique())
    labels.sort()
    for i in labels:
        tmp_data = data[data.label_cartography == i]
        plt.scatter(tmp_data.p,
                    tmp_data.z,
                    label=i, **args_dot)
    plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left', borderaxespad=0)

    if not gois is None:
        for i in gois:
            __plot_goi(data, i, args_annot)

def plot_cartography_kde(data_, gois=None, scatter=False, kde=True,
                             args_kde={},
                             args_line={},
                             args_annot={}):
    """

    kind: str
        "kde", "hex"
    """
    default = {"cmap": "RdPu",
               "shade":True,
               "shade_lowest":False,
               "n_levels": 12}#"RdPu"
    default.update(args_kde)
    args_kde = default.copy()

    default = {"alpha": 0.5, "c":"white"}
    default.update(args_line)
    args_line = default.copy()

    if not args_kde["shade"]:
        args_line.update({"c": "gray"})


    data = data_.copy()
    data = data.dropna(axis=0)

    data.columns = ["z", "p"]
    #data = data[(data.p>=0) & (data.z >= 0)]

    data = add_label_cartography(data)

    _plot_base_cartography(z_min=-1, z_max=data.z.max(), args_line=args_line)

    labels = list(data.label_cartography.unique())
    labels.sort()

    x, y = data.p, data.z,
    if kde:
        sns.kdeplot(x, y, **args_kde)

    if scatter:
        plt.scatter(x, y,  marker='o', edgecolor="lightgray",
                    c="none", alpha=1)
    #plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left', borderaxespad=0)

    if not gois is None:
        gois = np.intersect1d(data.index.values, gois)
        for i in gois:
            __plot_goi(data, i, args_annot, True)
    plt.xlabel("Participation coefficient (P)")
    plt.ylabel("Whithin-module\ndegree (z)")



# -*- coding: utf-8 -*-
'''
This is a series of custom functions for the inferring of GRN from single cell RNA-seq data.

Codes were written by Kenji Kamimoto.


'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing
import numpy as np
import pandas as pd

# 0.2. libraries for Machine learning modeling
from sklearn.ensemble import BaggingRegressor
from sklearn.linear_model import Ridge, BayesianRidge


# 0.3. custom libraries
from ..utility import intersect

###############################
### bayesian_ridge function ###
###############################

def get_bayesian_ridge_coefs(target_gene, gem, gem_scaled, TFdict, cellstate=None,
                             scaling=True):

    #print(target_gene)
    ## 1. Data prep
    if target_gene not in TFdict.keys():
        #print("err")
        return 0, 0, 0

    # define regGenes
    reggenes = TFdict[target_gene]
    allgenes_detected = list(gem.columns)
    reggenes = intersect(reggenes, allgenes_detected)

    if target_gene in reggenes:
        reggenes.remove(target_gene)

    reg_all = reggenes.copy()

    if not reggenes: # if reqgene is empty, return 0
        return 0, 0, 0

    # add cell state data
    if not cellstate is None:
        cellstate_name = list(cellstate.columns)
        reg_all += cellstate_name


    # prepare learning data
    if scaling:
        data = gem_scaled[reg_all]
    else:
        data = gem[reg_all]
    label = gem[target_gene]

    # make model and fitting
    model_br = BayesianRidge()
    model_br.fit(data.values, label)

    coef_names = reggenes
    ind_ = np.where([i in reggenes for i in reg_all])

    coef_mean = model_br.coef_[ind_]
    coef_variance = model_br.sigma_.diagonal()[ind_]

    return coef_mean, coef_variance, coef_names

##############################
### bagging_ridge function ###
##############################

def get_bagging_ridge_coefs(target_gene, gem, gem_scaled, TFdict, cellstate=None,
                 bagging_number=1000, scaling=True, n_jobs=-1, alpha=1, solver="auto"):
    #print(target_gene)
    ## 1. Data prep
    if target_gene not in TFdict.keys():
        #print("err")
        return 0

    # define regGenes
    reggenes = TFdict[target_gene]
    allgenes_detected = list(gem.columns)
    reggenes = intersect(reggenes, allgenes_detected)

    if target_gene in reggenes:
        reggenes.remove(target_gene)

    reg_all = reggenes.copy()

    if not reggenes: # if reqgene is empty, return 0
        return 0

    # add cell state data
    if not cellstate is None:
        cellstate_name = list(cellstate.columns)
        reg_all += cellstate_name


    # prepare learning data
    if scaling:
        data = gem_scaled[reg_all]
    else:
        data = gem[reg_all]
    label = gem[target_gene]

    #print(n_jobs)
    # bagging model
    model = BaggingRegressor(base_estimator=Ridge(alpha=alpha,
                                                  solver=solver,
                                                  random_state=123),
                             n_estimators=bagging_number,
                             bootstrap=True,
                             max_features=0.8,
                             n_jobs=n_jobs,
                             verbose=False,
                             random_state=123)
    model.fit(data, label)

    # get results
    coefs = _get_coef_matrix(model, reg_all)

    # remove cellstate data from coefs
    coefs = coefs[intersect(reggenes, coefs.columns.values)]

    return coefs


# this is a function to extract coef information from sklearn ensemble_model.
def _get_coef_matrix(ensemble_model, feature_names):
    # ensemble_model: trained ensemble model. e.g. BaggingRegressor
    # feature_names: list or numpy array of feature names. e.g. feature_names=X_train.columns
    feature_names = np.array(feature_names)
    n_estimater = len(ensemble_model.estimators_features_)
    coef_list = \
        [pd.Series(ensemble_model.estimators_[i].coef_,
                   index=feature_names[ensemble_model.estimators_features_[i]])\
         for i in range(n_estimater)]

    coef_df = pd.concat(coef_list, axis=1, sort=False).transpose()

    return coef_df



# -*- coding: utf-8 -*-
import pandas as pd
from scanpy import AnnData

import celloracle as co

# Mouse scATAC base GRN loading function
def test_load_mouse_scATAC_atlas_base_GRN():
    df = co.data.load_mouse_scATAC_atlas_base_GRN(force_download=False)
    assert isinstance(df, pd.core.frame.DataFrame)



# -*- coding: utf-8 -*-
"""
The :mod:`.network` module implements GRN inference.

"""
from .net_core import Net
from .net_util import load_net_from_patquets, getDF_TGxTF, getDF_peakxTF, load_net


__all__ = ["Net",
           "load_net_from_patquets",
           "getDF_TGxTF",
           "getDF_peakxTF"
           "load_net"]



# -*- coding: utf-8 -*-
'''
This file contains custom functions for the analysis of ATAC-seq data.
Genomic activity information (peak of ATAC-seq) will be extracted first.
Then the peak DNA sequence will be subjected to TF motif scan.
Finally we will get list of TFs that potentially binds to a specific gene.

Codes were written by Kenji Kamimoto.


'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing

import pandas as pd
import numpy as np

import sys, os

from tqdm.auto import tqdm

# 0.2. libraries for DNA and genome data wrangling and Motif analysis
from genomepy import Genome

#from gimmemotifs.motif import Motif
from gimmemotifs.scanner import Scanner
from gimmemotifs.fasta import Fasta

from pybedtools import BedTool


####
### bed f
def decompose_chrstr(peak_str):
    """
    Take peak name as input and return splitted strs.

    Args:
        peak_str (str): peak name.

    Returns:
        tuple: splitted peak name.

    Examples:
       >>> decompose_chrstr("chr1_111111_222222")
       "chr1", "111111", "222222"
    """
    *chr_, start, end = peak_str.split("_")
    chr_ = "_".join(chr_)

    return chr_, start, end

def list_peakstr_to_df(x):
    """
    Convert list of peaks(str) into data frame.

    Args:
       x (list of str): list of peak names

    Returns:
       pandas.dataframe: peak info as DataFrame

    Examples:
       >>> x = ['chr1_3094484_3095479', 'chr1_3113499_3113979', 'chr1_3119478_3121690']
       >>> list_peakstr_to_df(x)
                   chr	start	end
            0	chr1	3094484	3095479
            1	chr1	3113499	3113979
            2	chr1	3119478	3121690
    """
    df = np.array([decompose_chrstr(i) for i in x])
    df = pd.DataFrame(df, columns=["chr", "start", "end"])
    df["start"] = df["start"].astype(int)
    df["end"] = df["end"].astype(int)
    return df

def df_to_list_peakstr(x):

    x = x.rename(columns={"chrom":"chr"})

    peak_str = x.chr + "_" + x.start.astype(str) + "_" + x.end.astype(str)
    peak_str = peak_str.values

    return peak_str

####
###



def peak_M1(peak_id):
    """
    Take a peak_id (index of bed file) as input,
    then subtract 1 from Start position.

    Args:
        peak_id (str): Index of bed file. It should be made of "chromosome name", "start position", "end position"
            e.g. "chr11_123445555_123445577"
    Returns:
        str: Processed peak_id.

    Examples:
        >>> a = "chr11_123445555_123445577"
        >>> peak_M1(a)
        "chr11_123445554_123445577"
    """
    chr_, start, end = decompose_chrstr(peak_id)
    return chr_ + "_" + str(int(start)-1) + "_" + end


def peak2fasta(peak_ids, ref_genome):

    '''
    Convert peak_id into fasta object.

    Args:
        peak_id (str or list of str): Peak_id.  e.g. "chr5_0930303_9499409"
            or it can be a list of peak_id.  e.g. ["chr5_0930303_9499409", "chr11_123445555_123445577"]

        ref_genome (str): Reference genome name.   e.g. "mm9", "mm10", "hg19" etc

    Returns:
        gimmemotifs fasta object: DNA sequence in fasta format

    '''
    genome_data = Genome(ref_genome)

    def peak2seq(peak_id):
        chromosome_name, start, end = decompose_chrstr(peak_id)
        locus = (int(start),int(end))

        tmp = genome_data[chromosome_name][locus[0]:locus[1]]
        name = f"{tmp.name}_{tmp.start}_{tmp.end}"
        seq = tmp.seq
        return (name, seq)


    if type(peak_ids) is str:
        peak_ids = [peak_ids]

    fasta = Fasta()
    for peak_id in peak_ids:
        name, seq = peak2seq(peak_id)
        fasta.add(name, seq)

    return fasta

def remove_zero_seq(fasta_object):
    """
    Remove DNA sequence with zero length
    """
    fasta = Fasta()
    for i, seq in enumerate(fasta_object.seqs):
        if seq:
            name = fasta_object.ids[i]
            fasta.add(name, seq)
    return fasta


def read_bed(bed_path):
    """
    Load bed file and return as dataframe.

    Args:
        bed_path (str): File path.

    Returns:
        pandas.dataframe: bed file in dataframe.

    """
    tt = BedTool(bed_path).to_dataframe().dropna(axis=0)
    tt["seqname"] = tt.chrom + "_" + tt.start.astype("int").astype("str") + "_" + tt.end.astype("int").astype("str")
    return tt
